###############################################################################
#
# Livewallpaper
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
#
###############################################################################
#
# This module provides the following macros:
# (inspired by CompizPlugin.cmake)
#
# livewallpaper_plugin(<plugin name>
#                      SOURCES source1 source2 ...
#                      GSCHEMA schema
#                      [PKGDEPS dep1 dep2 ...]
#                      [LDFLAGSADD flag1 flag2 ...]
#                      [CFLAGSADD flag1 flag2 ...]
#                      [LIBRARIES flag1 flag2 ...]
#                      [LIBDIRS dir1 dir2 ...]
#                      [INCDIRS dir1 dir2 ...])
#
# SOURCES    = all source files of the plugin
# GSCHEMA    = the plugin's *.gschema.xml file
# PKGDEPS    = pkgconfig dependencies
# LDFLAGSADD = flags added to the link command
# CFLAGSADD  = flags added to the compile command
# LIBRARIES  = libraries added to the link command
# LIBDIRS    = additional link directories
# INCDIRS    = additional include directories
#
# This macro compiles and installs the plugin and its settings.
#
#
# To install additional data files use:
#
# livewallpaper_plugin_install_data(<plugin name>
#                                   FILES file1 file2 ...
#                                   [OTHER PARAMETERS])
#
# or for directories:
#
# livewallpaper_plugin_install_data(<plugin name>
#                                   DIRECTORY dir1 dir2 ...
#                                   [OTHER PARAMETERS])
#
# This macro is only a wrapper for cmake's install command, so the documentation
# of all OTHER PARAMETERS can be found here:
#     http://www.cmake.org/cmake/help/v2.8.0/cmake.html#command:install
# You can use every parameter specified there except the DESTINATION parameter.
# The DESTINATION will always be the data dir of the given plugin.
#
#
# The following variables will be used by those macros:
#
# LIVEWALLPAPER_PLUGIN_INSTALL_TYPE = (package | livewallpaper | local (default))
#     package       = install into ${CMAKE_INSTALL_PREFIX}
#     livewallpaper = install into livewallpaper prefix
#                     (as specified in the livewallpaper.pc file)
#     local         = install into home directory
#
###############################################################################

find_package(Livewallpaper REQUIRED)

find_program(GLIB_COMPILE_SCHEMAS_EXECUTABLE glib-compile-schemas)

if(NOT GLIB_COMPILE_SCHEMAS_EXECUTABLE)
	message(FATAL_ERROR "glib-compile-schemas not found")
endif(NOT GLIB_COMPILE_SCHEMAS_EXECUTABLE)

mark_as_advanced(GLIB_COMPILE_SCHEMAS_EXECUTABLE)


macro(_init_directories)
	if(LIVEWALLPAPER_PLUGIN_INSTALL_TYPE)
		set(
			LIVEWALLPAPER_PLUGIN_INSTALL_TYPE ${LIVEWALLPAPER_PLUGIN_INSTALL_TYPE} CACHE STRING
			"Where a plugin should be installed \(package | livewallpaper | local\)"
		)
	else(LIVEWALLPAPER_PLUGIN_INSTALL_TYPE)
		set(
			LIVEWALLPAPER_PLUGIN_INSTALL_TYPE "local" CACHE STRING
			"Where a plugin should be installed \(package | livewallpaper | local\)"
		)
	endif(LIVEWALLPAPER_PLUGIN_INSTALL_TYPE)

	if("${LIVEWALLPAPER_PLUGIN_INSTALL_TYPE}" STREQUAL "package")
		set(PLUGIN_LIBDIR ${CMAKE_INSTALL_PREFIX}/lib/livewallpaper/plugins)
		set(PLUGIN_DATADIR ${CMAKE_INSTALL_PREFIX}/share/livewallpaper/plugins)
		set(PLUGIN_SCHEMA_DIR ${CMAKE_INSTALL_PREFIX}/share/glib-2.0/schemas)
	elseif("${LIVEWALLPAPER_PLUGIN_INSTALL_TYPE}" STREQUAL "livewallpaper")
		set(PLUGIN_LIBDIR ${LIVEWALLPAPER_PREFIX}/lib/livewallpaper/plugins)
		set(PLUGIN_DATADIR ${LIVEWALLPAPER_PREFIX}/share/livewallpaper/plugins)
		set(PLUGIN_SCHEMA_DIR ${LIVEWALLPAPER_PREFIX}/share/glib-2.0/schemas)
	else("${LIVEWALLPAPER_PLUGIN_INSTALL_TYPE}" STREQUAL "livewallpaper")
		set(PLUGIN_LIBDIR $ENV{HOME}/.livewallpaper/plugins)
		set(PLUGIN_DATADIR $ENV{HOME}/.livewallpaper/plugins)
		set(PLUGIN_SCHEMA_DIR $ENV{HOME}/.local/share/glib-2.0/schemas)
	endif("${LIVEWALLPAPER_PLUGIN_INSTALL_TYPE}" STREQUAL "package")
endmacro(_init_directories)

macro(_parse_arguments _prefix)
	set(_current_var _foo)
	set(_supported_var SOURCES GSCHEMA PKGDEPS LDFLAGSADD CFLAGSADD LIBRARIES LIBDIRS INCDIRS)
	foreach(_val ${_supported_var})
		set(${_prefix}_${_val})
	endforeach(_val)

	foreach(_val ${ARGN})
		set(_found FALSE)
		foreach(_find ${_supported_var})
			if("${_find}" STREQUAL "${_val}")
				set(_found TRUE)
			endif("${_find}" STREQUAL "${_val}")
		endforeach(_find)

		if(_found)
			set(_current_var ${_prefix}_${_val})
		else(_found)
			list(APPEND ${_current_var} ${_val})
		endif(_found)
	endforeach(_val)
endmacro(_parse_arguments)

macro(livewallpaper_plugin plugin)
	string(TOUPPER ${plugin} _PLUGIN)

	_parse_arguments(${_PLUGIN} ${ARGN})
	_init_directories()

	# check additional dependencies
	if(${_PLUGIN}_PKGDEPS)
		find_package(PkgConfig REQUIRED)

		pkg_check_modules(_${plugin}_deps REQUIRED ${${_PLUGIN}_PKGDEPS})

		list(APPEND ${_PLUGIN}_CFLAGSADD ${_${plugin}_deps_CFLAGS})
		list(APPEND ${_PLUGIN}_LDFLAGSADD ${_${plugin}_deps_LDFLAGS})
		list(APPEND ${_PLUGIN}_LIBRARIES ${_${plugin}_deps_LIBRARIES})
		list(APPEND ${_PLUGIN}_LIBDIRS ${_${plugin}_deps_LIBRARY_DIRS})
		list(APPEND ${_PLUGIN}_INCDIRS ${_${plugin}_deps_INCLUDE_DIRS})
	endif(${_PLUGIN}_PKGDEPS)


	# compile plugin
	include_directories(
		${${_PLUGIN}_INCDIRS}
		${LIVEWALLPAPER_INCLUDEDIR}
		${LIVEWALLPAPER_INCLUDE_DIRS}
	)

	link_directories(
		${${_PLUGIN}_LIBDIRS}
		${LIVEWALLPAPER_LIBDIR}
		${LIVEWALLPAPER_LIBRARY_DIRS}
	)

	add_library(
		${plugin} SHARED ${${_PLUGIN}_SOURCES}
	)

	install(
		TARGETS ${plugin}
		DESTINATION ${PLUGIN_LIBDIR}/${plugin}
	)

	target_link_libraries(
		${plugin}
		${${_PLUGIN}_LIBRARIES}
		${LIVEWALLPAPER_LIBRARIES}
	)

	# set cflags and ldflags
	set(_${plugin}_properties)
	foreach(_flag ${${_PLUGIN}_CFLAGSADD} ${LIVEWALLPAPER_CFLAGS})
		list(APPEND _${plugin}_properties COMPILE_FLAGS ${_flag})
	endforeach(_flag)
	foreach(_flag ${${_PLUGIN}_LDFLAGSADD} ${LIVEWALLPAPER_LDFLAGS})
		list(APPEND _${plugin}_properties LINK_FLAGS ${_flag})
	endforeach(_flag)

	set_target_properties(
		${plugin} PROPERTIES
		${_${plugin}_properties}
	)


	# install .plugin file
	find_file(
		_${plugin}_plugin_in ${plugin}.plugin.in
		PATHS ${CMAKE_CURRENT_SOURCE_DIR}
		NO_DEFAULT_PATH
	)

	if(_${plugin}_plugin_in)
		set(_${plugin}_plugin ${CMAKE_CURRENT_BINARY_DIR}/${plugin}.plugin)
		configure_file(
			${_${plugin}_plugin_in}
			${_${plugin}_plugin}
		)
	else(_${plugin}_plugin_in)
		find_file(
			_${plugin}_plugin ${plugin}.plugin
			PATHS ${CMAKE_CURRENT_SOURCE_DIR}
			NO_DEFAULT_PATH
		)
	endif(_${plugin}_plugin_in)

	mark_as_advanced(_${plugin}_plugin _${plugin}_plugin_in)

	if(_${plugin}_plugin)
		install(
			FILES ${_${plugin}_plugin}
			DESTINATION ${PLUGIN_LIBDIR}/${plugin}
		)
	endif(_${plugin}_plugin)


	# install gschema
	install(
		FILES ${${_PLUGIN}_GSCHEMA}
		DESTINATION ${PLUGIN_SCHEMA_DIR}
	)

	install(CODE "execute_process(COMMAND ${GLIB_COMPILE_SCHEMAS_EXECUTABLE} ${PLUGIN_SCHEMA_DIR})")
endmacro(livewallpaper_plugin)

macro(livewallpaper_plugin_install_data plugin)
	_init_directories()

	install(${ARGN} DESTINATION ${PLUGIN_DATADIR}/${plugin})
endmacro(livewallpaper_plugin_install_data)

