/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <string.h>
#include <livewallpaper/core.h>


#define LW_BACKGROUND_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_TYPE_BACKGROUND, LwBackgroundPrivate))

struct _LwBackgroundPrivate
{
	LwTexture *default_bg;
	LwBackgroundRenderType default_render_type;


	gboolean use_custom;
	LwTexture *custom_bg;
	LwBackgroundRenderType custom_render_type;

	gchar *image;
	LwBackgroundRenderType render_type;
	GdkRGBA primary_color;
	GdkRGBA secondary_color;
	LwBackgroundShadingType shading_type;
};

enum
{
	PROP_0,

	PROP_USE_CUSTOM_BG,
	PROP_IMAGE,
	PROP_RENDER_TYPE,
	PROP_PRIMARY_COLOR,
	PROP_SECONDARY_COLOR,
	PROP_SHADING_TYPE,

	N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(LwBackground, lw_background, G_TYPE_OBJECT)

/**
 * lw_background_new_with_default_image:
 * @filename: Name of the default background image
 * @type: #LwBackgroundRenderType to use for default background image
 * @error: Return location of an error
 *
 * Creates a new #LwBackground with an default background image.
 *
 * Returns: A new #LwBackground
 */
LwBackground*
lw_background_new_with_default_image(const char *filename,
                                     LwBackgroundRenderType type,
                                     GError **error)
{
	LwBackground *background = g_object_new(LW_TYPE_BACKGROUND, NULL);

	background->priv->default_bg = lw_texture_new_from_file(filename, error);
	background->priv->default_render_type = type;

	return background;
}

static LwTexture*
create_texture_from_color(GdkColor *primary_color,
                          GdkColor *secondary_color,
                          LwBackgroundShadingType type)
{
	guint width, height;
	GLushort data[8];

	data[0] = primary_color->red;
	data[1] = primary_color->green;
	data[2] = primary_color->blue;
	data[3] = 65535;

	data[4] = secondary_color->red;
	data[5] = secondary_color->green;
	data[6] = secondary_color->blue;
	data[7] = 65535;

	switch(type)
	{
		case LwBackgroundHorizontal:
			width = 2;
			height = 1;
			break;

		case LwBackgroundVertical:
			width = 1;
			height = 2;
			break;

		case LwBackgroundSolid:
		default:
			width = 1;
			height = 1;
	}

	return lw_texture_new_from_data((guchar *) data, width, height, GL_RGBA, GL_UNSIGNED_SHORT);
}

static LwTexture*
create_texture_from_rgba(GdkRGBA *primary_rgba,
                         GdkRGBA *secondary_rgba,
                         LwBackgroundShadingType type)
{
	GdkColor primary_color, secondary_color;
	primary_color.red = primary_rgba->red * 65535;
	primary_color.green = primary_rgba->green * 65535;
	primary_color.blue = primary_rgba->blue * 65535;

	secondary_color.red = secondary_rgba->red * 65535;
	secondary_color.green = secondary_rgba->green * 65535;
	secondary_color.blue = secondary_rgba->blue * 65535;

	return create_texture_from_color(&primary_color, &secondary_color, type);
}

/**
 * lw_background_new_with_default_color:
 * @primary_color: Left or bottom color when drawing gradients, or the solid color
 * @secondary_color: Right or top color when drawing gradients, not used for solid color
 * @type: The type of color gradient
 *
 * Creates a new #LwBackground with an color gradient as default background.
 *
 * Returns: A new #LwBackground
 */
LwBackground*
lw_background_new_with_default_color(GdkColor *primary_color,
                                     GdkColor *secondary_color,
                                     LwBackgroundShadingType type)
{
	LwBackground *background = g_object_new(LW_TYPE_BACKGROUND, NULL);

	background->priv->default_bg = create_texture_from_color(primary_color, secondary_color, type);
	background->priv->default_render_type = LwBackgroundStretched;

	lw_texture_set_filter(background->priv->default_bg, GL_LINEAR);

	return background;
}

static void
lw_background_update_custom_bg(LwBackground *self)
{
	if(self->priv->image != NULL)
	{
		GError *error = NULL;

		/* Load image */
		if(self->priv->custom_bg != NULL)
			g_object_unref(self->priv->custom_bg);

		self->priv->custom_bg = lw_texture_new_from_file(self->priv->image, &error);
		if(error != NULL)
		{
			if(strlen(self->priv->image) != 0)
				g_warning("Could not load custom background image: %s", error->message);
			g_error_free(error);

			/* Use color gradient as background */
			g_free(self->priv->image);
			self->priv->image = NULL;
			lw_background_update_custom_bg(self);
			return;
		}

		self->priv->custom_render_type = self->priv->render_type;
	}
	else
	{
		/* Create color gradient */
		if(self->priv->custom_bg != NULL)
			g_object_unref(self->priv->custom_bg);

		self->priv->custom_bg = create_texture_from_rgba(&self->priv->primary_color,
		                                                 &self->priv->secondary_color,
		                                                 self->priv->shading_type);
		self->priv->custom_render_type = LwBackgroundStretched;

		lw_texture_set_filter(self->priv->custom_bg, GL_LINEAR);
	}
}

static void
lw_background_set_property(GObject *object,
                           guint property_id,
                           const GValue *value,
                           GParamSpec *pspec)
{
	LwBackground *self = LW_BACKGROUND(object);

	switch(property_id)
	{
		case PROP_USE_CUSTOM_BG:
			self->priv->use_custom = g_value_get_boolean(value);
			break;

		case PROP_IMAGE:
			g_free(self->priv->image);
			self->priv->image = g_strdup(g_value_get_string(value));
			lw_background_update_custom_bg(self);
			break;

		case PROP_RENDER_TYPE:
			self->priv->render_type = g_value_get_uint(value);
			if(self->priv->image != NULL)
				self->priv->custom_render_type = self->priv->render_type;
			break;

		case PROP_PRIMARY_COLOR:
			self->priv->primary_color = *((GdkRGBA*) g_value_get_boxed(value));
			if(self->priv->image == NULL)
				lw_background_update_custom_bg(self);
			break;

		case PROP_SECONDARY_COLOR:
			self->priv->secondary_color = *((GdkRGBA*) g_value_get_boxed(value));
			if(self->priv->image == NULL)
				lw_background_update_custom_bg(self);
			break;

		case PROP_SHADING_TYPE:
			self->priv->shading_type = g_value_get_uint(value);
			if(self->priv->image == NULL)
				lw_background_update_custom_bg(self);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
lw_background_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
	LwBackground *self = LW_BACKGROUND(object);

	switch(property_id)
	{
		case PROP_USE_CUSTOM_BG:
			g_value_set_boolean(value, self->priv->use_custom);
			break;

		case PROP_IMAGE:
			g_value_set_string(value, self->priv->image);
			break;

		case PROP_RENDER_TYPE:
			g_value_set_uint(value, self->priv->render_type);
			break;

		case PROP_PRIMARY_COLOR:
			g_value_set_boxed(value, &self->priv->primary_color);
			break;

		case PROP_SECONDARY_COLOR:
			g_value_set_boxed(value, &self->priv->secondary_color);
			break;

		case PROP_SHADING_TYPE:
			g_value_set_uint(value, self->priv->shading_type);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

/**
 * lw_background_draw:
 * @self: A #LwBackground
 * @output: The active output
 *
 * Draws the background to the active output.
 */
void
lw_background_draw(LwBackground *self, LwOutput *output)
{
	/* Use gint instead of guint! */
	gint width  = lw_output_get_width(output),
	     height = lw_output_get_height(output);
	gint tex_width, tex_height;
	LwTexture* tex;
	LwBackgroundRenderType type;
	LwTextureMatrix m;

	/* Get the right background */
	if(self->priv->use_custom)
	{
		tex = self->priv->custom_bg;
		type = self->priv->custom_render_type;
	}
	else
	{
		tex = self->priv->default_bg;
		type = self->priv->default_render_type;
	}

	if(tex == NULL) return;

	tex_width = lw_texture_get_width(tex);
	tex_height = lw_texture_get_height(tex);

	/* Save and reset the coordinate system for drawing the background. */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(0, width, height, 0, -1.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	/* Apply render type */
	m = tex->matrix;
	switch(type)
	{
		case LwBackgroundCentered:
			m.x0 -= ((width - tex_width) / 2) * m.xx;
			m.y0 -= ((height - tex_height) / 2) * m.yy;

			lw_texture_set_wrap(tex, GL_CLAMP_TO_EDGE);
			break;

		case LwBackgroundScaled:
		{
			float s = MIN((float) width / tex_width,
			              (float) height / tex_height);

			m.xx /= s;
			m.yy /= s;

			m.x0 -= ((width - (tex_width * s)) / 2) * m.xx;
			m.y0 -= ((height - (tex_height * s)) / 2) * m.yy;

			lw_texture_set_wrap(tex, GL_CLAMP_TO_EDGE);
			break;
		}

		case LwBackgroundStretched:
			m.xx /= (float) width / tex_width;
			m.yy /= (float) height / tex_height;

			lw_texture_set_wrap(tex, GL_CLAMP_TO_EDGE);
			break;

		case LwBackgroundTiled:
			lw_texture_set_wrap(tex, GL_REPEAT);
			break;

		case LwBackgroundZoom:
		default:
		{
			float s = MAX((float) width / tex_width,
			              (float) height / tex_height);

			m.xx /= s;
			m.yy /= s;

			m.x0 -= ((width - (tex_width * s)) / 2) * m.xx;
			m.y0 -= ((height - (tex_height * s)) / 2) * m.yy;

			lw_texture_set_wrap(tex, GL_CLAMP_TO_EDGE);
		}
	}

	/* Draw background */
	glColor3f(1.0f, 1.0f, 1.0f);
	lw_texture_enable(tex);

	glBegin(GL_QUADS);
		glTexCoord2f(
			LW_TEX_COORD_X(m, 0.0f),
			LW_TEX_COORD_Y(m, 0.0f)
		);
		glVertex2f(0.0f, 0.0f);

		glTexCoord2f(
			LW_TEX_COORD_X(m, 0.0f),
			LW_TEX_COORD_Y(m, height)
		);
		glVertex2f(0.0f, height);

		glTexCoord2f(
			LW_TEX_COORD_X(m, width),
			LW_TEX_COORD_Y(m, height)
		);
		glVertex2f(width, height);

		glTexCoord2f(
			LW_TEX_COORD_X(m, width),
			LW_TEX_COORD_Y(m, 0.0f)
		);
		glVertex2f(width, 0.0f);
	glEnd();

	lw_texture_disable(tex);

	/* Restore the coordinate system. */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

static void
lw_background_bg_render_type_changed(GSettings *settings, gchar *key, LwBackground *self)
{
	(void) key;

	g_object_set(self, "render-type", g_settings_get_enum(settings, "bg-render-type"), NULL);
}

static void
lw_background_bg_shading_type_changed(GSettings *settings, gchar *key, LwBackground *self)
{
	(void) key;

	g_object_set(self, "shading-type", g_settings_get_enum(settings, "bg-shading-type"), NULL);
}

static void
lw_background_init(LwBackground *self)
{
	GSettings *settings = g_settings_new("net.launchpad.livewallpaper");
	self->priv = LW_BACKGROUND_GET_PRIVATE(self);

	self->priv->default_bg = NULL;
	self->priv->default_render_type = LwBackgroundZoom;

	self->priv->use_custom = FALSE;
	self->priv->custom_bg = NULL;
	self->priv->custom_render_type = LwBackgroundZoom;

	self->priv->render_type = g_settings_get_enum(settings, "bg-render-type");
	self->priv->shading_type = g_settings_get_enum(settings, "bg-shading-type");

	/* Bind settings to properties */
	g_settings_bind(settings, "use-custom-bg", self, "use-custom-bg", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "bg-image", self, "image", G_SETTINGS_BIND_GET);
	g_signal_connect(settings, "changed::bg-render-type", G_CALLBACK(lw_background_bg_render_type_changed), self);

	g_settings_bind_with_mapping(settings, "bg-primary-color", self, "primary-color", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
	g_settings_bind_with_mapping(settings, "bg-secondary-color", self, "secondary-color", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
	g_signal_connect(settings, "changed::bg-shading-type", G_CALLBACK(lw_background_bg_shading_type_changed), self);
}

static void
lw_background_dispose(GObject *object)
{
	LwBackground *self = LW_BACKGROUND(object);

	if(self->priv->default_bg)
	{
		g_object_unref(self->priv->default_bg);
		self->priv->default_bg = NULL;
	}

	if(self->priv->custom_bg)
	{
		g_object_unref(self->priv->custom_bg);
		self->priv->custom_bg = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_background_parent_class)->dispose(object);
}

static void
lw_background_finalize(GObject *object)
{
	LwBackground *self = LW_BACKGROUND(object);
	
	g_free(self->priv->image);
	
	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_background_parent_class)->finalize(object);
}

static void
lw_background_class_init(LwBackgroundClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = lw_background_set_property;
	gobject_class->get_property = lw_background_get_property;
	gobject_class->dispose = lw_background_dispose;
	gobject_class->finalize = lw_background_finalize;

	g_type_class_add_private(klass, sizeof(LwBackgroundPrivate));

	/**
	 * LwBackground:use-custom-bg:
	 *
	 * Use custom background or the default background of the active plugin
	 */
	obj_properties[PROP_USE_CUSTOM_BG] =
		g_param_spec_boolean("use-custom-bg",
		                     "Use Custom Background",
		                     "Use custom background or the default background of the active plugin",
		                     FALSE,
		                     G_PARAM_READWRITE);

	/**
	 * LwBackground:image:
	 *
	 * Path to use for the background image
	 */
	obj_properties[PROP_IMAGE] =
		g_param_spec_string("image",
		                    "Image",
		                    "Path to use for the background image",
		                    NULL,
		                    G_PARAM_READWRITE);

	/**
	 * LwBackground:render-type:
	 *
	 * Determines how the background image is rendered
	 */
	obj_properties[PROP_RENDER_TYPE] =
		g_param_spec_uint("render-type",
		                  "Render Type",
		                  "Determines how the background image is rendered",
		                  0,
		                  4,
		                  0,
		                  G_PARAM_READWRITE);

	/**
	 * LwBackground:primary-color:
	 *
	 * Left or bottom color when drawing gradients, or the solid color
	 */
	obj_properties[PROP_PRIMARY_COLOR] =
		g_param_spec_boxed("primary-color",
		                   "Primary Color",
		                   "Left or bottom color when drawing gradients, or the solid color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	/**
	 * LwBackground:secondary-color:
	 *
	 * Right or top color when drawing gradients, not used for solid color
	 */
	obj_properties[PROP_SECONDARY_COLOR] =
		g_param_spec_boxed("secondary-color",
		                   "Secondary Color",
		                   "Right and top color when drawing gradients, not used for solid color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	/**
	 * LwBackground:shading-type:
	 *
	 * How to shade the background color
	 */
	obj_properties[PROP_SHADING_TYPE] =
		g_param_spec_uint("shading-type",
		                  "Shading Type",
		                  "How to shade the background color",
		                  0,
		                  2,
		                  0,
		                  G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

