/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 *
 * The color mapping functions are taken from livewallpaper-config. They
 * were originally inspirated by gnome-terminal (profile-editor.c).
 *
 * The color conversion functions are inspirated by GIMP:
 * http://git.gnome.org/browse/gimp/tree/libgimpcolor/gimpcolorspace.c?id=GIMP_2_6_11
 *
 */

#include <livewallpaper/core.h>

/**
 * lw_color_get_color_mapping:
 * @value: Return location for the property value
 * @variant: The GVariant
 * @p: User data that was specified when the binding was created
 *
 * This function is used to convert from GSettings to GdkRGBA. Use this function
 * as get mapping function of g_settings_bind_with_mapping.
 *
 * Returns: TRUE if the conversion succeeded, FALSE in case of an error
 */
gboolean
lw_color_get_color_mapping(GValue *value, GVariant *variant, gpointer p)
{
	const gchar *s;
	GdkRGBA rgba;
	(void) p;

	g_variant_get(variant, "&s", &s);
	if(!gdk_rgba_parse(&rgba, s)) return FALSE;

	g_value_set_boxed(value, &rgba);
	return TRUE;
}

/**
 * lw_color_set_color_mapping:
 * @value: A GValue containing the property value to map
 * @type: The GVariantType to create
 * @p: User data that was specified when the binding was created
 *
 * This function is used to convert a GdkRGBA to GSettings. Use this function as
 * set mapping function of g_settings_bind_with_mapping.
 *
 * Returns: A new GVariant holding the data from value, or NULL in case of an error
 */
GVariant*
lw_color_set_color_mapping(const GValue *value, const GVariantType *type, gpointer p)
{
	(void) p;
	(void) type;

	return g_variant_new_string( gdk_rgba_to_string(g_value_get_boxed(value)) );
}

/**
 * lw_rgb_to_hsl:
 * @red: A pointer to the red color component. Return location of the hue.
 * @green: A pointer to the green color component. Return location of the saturation.
 * @blue: A pointer to the blue color component. Return location of the lightness.
 *
 * Converts a color from the RGB color model to the HSL color model.
 */
void
lw_rgb_to_hsl(float *red, float *green, float *blue)
{
	float r, g, b;
	float h, s, l;
	float min, max;
	float delta;

	r = *red;
	g = *green;
	b = *blue;

	max = MAX(r, MAX(g, b));
	min = MIN(r, MIN(g, b));

	l = (max + min) / 2.0f;

	if(max == min)
	{
		s = 0.0f;
		h = 0.0f;
	}
	else
	{
		delta = max - min;

		if(l <= 0.5f)
			s = delta / (max + min);
		else
			s = delta / (2.0f - max - min);

		if(r == max)
			h = (g - b) / delta;
		else if(g == max)
			h = 2.0f + (b - r) / delta;
		else
			h = 4.0f + (r - g) / delta;

		h /= 6.0f;

		if(h < 0.0f)
			h += 1.0f;
		else if(h > 1.0f)
			h -= 1.0f;
	}

	*red = h;
	*green = s;
	*blue = l;
}

static inline float
hsl_value(float n1, float n2, float hue)
{
	if(hue < 0.0f)
		hue += 6.0f;
	else if(hue > 6.0f)
		hue -= 6.0f;

	if(hue < 1.0f)
		return n1 + (n2 - n1) * hue;
	else if(hue < 3.0f)
		return n2;
	else if(hue < 4.0f)
		return n1 + (n2 - n1) * (4.0f - hue);
	else
		return n1;
}

/**
 * lw_hsl_to_rgb:
 * @hue: A pointer to hue. Return location of the red color component.
 * @saturation: A pointer to saturation. Return location of the green color component.
 * @lightness: A pointer to lightness. Return location of the blue color component.
 *
 * Converts a color from the HSL color model to the RGB color model.
 */
void
lw_hsl_to_rgb(float *hue, float *saturation, float *lightness)
{
	float h, s, l;

	h = *hue;
	s = *saturation;
	l = *lightness;

	if(s == 0.0f)
	{
		*hue = l;
		*lightness = l;
		*saturation = l;
	}
	else
	{
		float m1, m2;

		/* Case 1 */
		if(l <= 0.5f)
			m2 = l * (1.0f + s);

		/* Case 2 */
		else
			m2 = l + s - l * s;

		m1 = 2.0 * l - m2;

		*hue = hsl_value(m1, m2, h * 6.0f + 2.0f);
		*saturation = hsl_value(m1, m2, h * 6.0f);
		*lightness = hsl_value(m1, m2, h * 6.0f - 2.0f);
	}
}

