/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <livewallpaper/core.h>

G_DEFINE_INTERFACE(LwLivewallpaperPlugin, lw_livewallpaper_plugin, G_TYPE_OBJECT)

/**
 * lw_livewallpaper_plugin_init_plugin:
 * @self: A #LwLivewallpaperPlugin
 *
 * Initialization of the plugin. Initialize all the variables of the plugin here that
 * need the plugin's data dir path.
 */
void
lw_livewallpaper_plugin_init_plugin(LwLivewallpaperPlugin *self)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->init_plugin)
		iface->init_plugin(self);
}

/**
 * lw_livewallpaper_plugin_adjust_viewport:
 * @self: A #LwLivewallpaperPlugin
 * @output: The new active #LwOutput
 *
 * Adjust the viewport to fit to this output. All following calls to lw_livewallpaper_plugin_paint
 * will get this output as parameter. This is the right place to transform the model or
 * projection matrix or to enable OpenGL feature's like blending. You should restore all
 * changes this function makes within the lw_livewallpaper_plugin_restore_viewport function.
 */
void
lw_livewallpaper_plugin_adjust_viewport(LwLivewallpaperPlugin *self, LwOutput *output)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->adjust_viewport)
		iface->adjust_viewport(self, output);
}

/**
 * lw_livewallpaper_plugin_prepare_paint:
 * @self: A #LwLivewallpaperPlugin
 * @ms_since_last_paint: The time since the last paint in milliseconds
 *
 * Update the animation for all outputs here.
 */
void
lw_livewallpaper_plugin_prepare_paint(LwLivewallpaperPlugin *self, gint ms_since_last_paint)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->prepare_paint)
		iface->prepare_paint(self, ms_since_last_paint);
}

/**
 * lw_livewallpaper_plugin_paint:
 * @self: A #LwLivewallpaperPlugin
 * @output: The #LwOutput to paint to
 *
 * Paint the wallpaper for the given output.
 */
void
lw_livewallpaper_plugin_paint(LwLivewallpaperPlugin *self, LwOutput *output)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->paint)
		iface->paint(self, output);
}

/**
 * lw_livewallpaper_plugin_done_paint:
 * @self: A #LwLivewallpaperPlugin
 *
 * Clean up after all outputs are painted.
 */
void
lw_livewallpaper_plugin_done_paint(LwLivewallpaperPlugin *self)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->done_paint)
		iface->done_paint(self);
}

/**
 * lw_livewallpaper_plugin_restore_viewport:
 * @self: A #LwLivewallpaperPlugin
 *
 * Restore all changes made by the lw_livewallpaper_plugin_adjust_viewport function
 * here.
 */
void
lw_livewallpaper_plugin_restore_viewport(LwLivewallpaperPlugin *self)
{
	LwLivewallpaperPluginInterface *iface;

	g_return_if_fail( LW_IS_LIVEWALLPAPER_PLUGIN(self) );

	iface = LW_LIVEWALLPAPER_PLUGIN_GET_INTERFACE(self);
	if(iface->restore_viewport)
		iface->restore_viewport(self);
}

static void
lw_livewallpaper_plugin_default_init(LwLivewallpaperPluginInterface *iface)
{
	static gboolean is_initialized = FALSE;

	if(!is_initialized)
	{
		(void) iface;
		is_initialized = TRUE;
	}
}

