/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 *
 * The LwTexture object is inspired by compiz' GLTexture class.
 *
 */

#include <livewallpaper/core.h>


static LwTextureMatrix identity_texture_matrix = {
	1.0f, 0.0f,
	0.0f, 1.0f,
	0.0f, 0.0f
};

#define LW_TEXTURE_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_TYPE_TEXTURE, LwTexturePrivate))

struct _LwTexturePrivate
{
	guint name;
	guint target;
	guint filter;
	guint wrap;

	gint width;
	gint height;
};

enum
{
	PROP_0,

	PROP_WIDTH,
	PROP_HEIGHT,

	N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(LwTexture, lw_texture, G_TYPE_OBJECT)

/**
 * lw_texture_new_from_file:
 * @filename: Name of the file to load
 * @error: Return location of an error
 *
 * Creates a new texture by loading an image from a file. This function supports
 * all file formats supported by gdk-pixbuf. If NULL is returned, then @error will
 * be set.
 *
 * Returns: A new #LwTexture, or NULL in case of an error
 */
LwTexture*
lw_texture_new_from_file(const char *filename, GError **error)
{
	LwTexture *texture;
	GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(filename, error);

	if(!pixbuf) return NULL;

	texture = lw_texture_new_from_pixbuf(pixbuf);

	g_object_unref(pixbuf);
	return texture;
}

/**
 * lw_texture_new_from_pixbuf:
 * @pixbuf: A GdkPixbuf holding the image data
 *
 * Creates a new texture by using the image data of the GdkPixbuf.
 *
 * Returns: A new #LwTexture
 */
LwTexture*
lw_texture_new_from_pixbuf(GdkPixbuf *pixbuf)
{
	guint width  = gdk_pixbuf_get_width(pixbuf),
		  height = gdk_pixbuf_get_height(pixbuf),
	      format = (gdk_pixbuf_get_has_alpha(pixbuf)) ? GL_RGBA : GL_RGB;
	guchar *data = gdk_pixbuf_get_pixels(pixbuf);

	return lw_texture_new_from_data(data, width, height, format, GL_UNSIGNED_BYTE);
}

/**
 * lw_texture_new_from_data:
 * @data: A pointer to the image data
 * @width: Width of the image in pixels
 * @height: Height of the image in pixels
 * @format: Format of the image data
 * @type: Data type of the image data
 *
 * Creates a new texture out of in-memory image data. Internally <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glTexImage2D.xml">glTexImage2D</ulink>
 * is used to create the texture. You can find a more detailed description of
 * @format and @type on the documentation page of <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glTexImage2D.xml">glTexImage2D</ulink>.
 *
 * Returns: A new #LwTexture
 */
LwTexture*
lw_texture_new_from_data(const guchar *data,
                         guint width,
                         guint height,
                         guint format,
                         guint type)
{
	LwTexture *texture;

	texture = g_object_new(LW_TYPE_TEXTURE,
	                       "width", width,
	                       "height", height,
	                       NULL);

	glBindTexture(lw_texture_get_target(texture),
	              lw_texture_get_name(texture));

	glTexImage2D(lw_texture_get_target(texture),
	             0,
	             GL_RGBA,
	             width, height,
	             0,
	             format,
	             type,
	             data);

	lw_texture_set_filter(texture, GL_NEAREST);
	lw_texture_set_wrap(texture, GL_CLAMP_TO_EDGE);

	return texture;
}

static void
lw_texture_set_property(GObject *object,
                        guint property_id,
                        const GValue *value,
                        GParamSpec *pspec)
{
	LwTexture *self = LW_TEXTURE(object);

	switch(property_id)
	{
		case PROP_WIDTH:
			self->priv->width = g_value_get_uint(value);
			self->matrix.xx = 1.0f / self->priv->width;
			break;

		case PROP_HEIGHT:
			self->priv->height = g_value_get_uint(value);
			self->matrix.yy = 1.0f / self->priv->height;
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
lw_texture_get_property(GObject *object,
                        guint property_id,
                        GValue *value,
                        GParamSpec *pspec)
{
	LwTexture *self = LW_TEXTURE(object);

	switch(property_id)
	{
		case PROP_WIDTH:
			g_value_set_uint(value, lw_texture_get_width(self));
			break;

		case PROP_HEIGHT:
			g_value_set_uint(value, lw_texture_get_height(self));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

/**
 * lw_texture_get_name:
 * @self: A #LwTexture
 *
 * Returns: The texture name returned by glGenTextures
 */
guint
lw_texture_get_name(LwTexture *self)
{
	return self->priv->name;
}

/**
 * lw_texture_get_target:
 * @self: A #LwTexture
 *
 * Returns the texture target. At the moment only GL_TEXTURE_2D is supported, so this function
 * always returns GL_TEXTURE_2D.
 *
 * Returns: The texture target
 */
guint
lw_texture_get_target(LwTexture *self)
{
	return self->priv->target;
}

/**
 * lw_texture_set_filter:
 * @self: A #LwTexture
 * @filter: The new texture filter
 *
 * Sets the texture's minifying and magnification function. Possible values are GL_NEAREST and
 * GL_LINEAR. GL_NEAREST is generally faster than GL_LINEAR, but it produces sharper edges.
 * #LwTexture uses GL_NEAREST by default.
 *
 * To get a more detailed description of all possible filters, take a look at the documentation
 * page of <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glTexParameter.xml">glTexParameter</ulink>.
 * This functions sets the GL_TEXTURE_MIN_FILTER and GL_TEXTURE_MAG_FILTER at the same time, so
 * only filters supported by both are vaild arguments.
 */
void
lw_texture_set_filter(LwTexture *self, guint filter)
{
	glBindTexture(self->priv->target, self->priv->name);

	self->priv->filter = filter;

	glTexParameteri(self->priv->target, GL_TEXTURE_MIN_FILTER, filter);
	glTexParameteri(self->priv->target, GL_TEXTURE_MAG_FILTER, filter);

	glBindTexture(self->priv->target, 0);
}

/**
 * lw_texture_get_filter:
 * @self: A #LwTexture
 *
 * Returns: The currently applied texture filter
 */
guint
lw_texture_get_filter(LwTexture *self)
{
	return self->priv->filter;
}

/**
 * lw_texture_set_wrap:
 * @self: A #LwTexture
 * @wrap: The new wrap parameter
 *
 * Sets the wrap parameter for texture coordinate s and t. The wrap parameter is set
 * to GL_CLAMP_TO_EDGE by default.
 *
 * This functions calls <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glTexParameter.xml">glTexParameter</ulink>
 * for GL_TEXTURE_WRAP_S and GL_TEXTURE_WRAP_T. You can find a complete list of all supported
 * parameters at the documentation page of <ulink url="http://www.opengl.org/sdk/docs/man/xhtml/glTexParameter.xml">glTexParameter</ulink>.
 */
void
lw_texture_set_wrap(LwTexture *self, guint wrap)
{
	glBindTexture(self->priv->target, self->priv->name);

	self->priv->wrap = wrap;

	glTexParameteri(self->priv->target, GL_TEXTURE_WRAP_S, wrap);
	glTexParameteri(self->priv->target, GL_TEXTURE_WRAP_T, wrap);

	glBindTexture(self->priv->target, 0);
}

/**
 * lw_texture_get_wrap:
 * @self: A #LwTexture
 *
 * Returns: The current wrap parameter
 */
guint
lw_texture_get_wrap(LwTexture *self)
{
	return self->priv->wrap;
}

/**
 * lw_texture_get_width:
 * @self: A #LwTexture
 *
 * Returns: The width of the texture
 */
guint
lw_texture_get_width(LwTexture *self)
{
	return self->priv->width;
}

/**
 * lw_texture_get_height:
 * @self: A #LwTexture
 *
 * Returns: The height of the texture
 */
guint
lw_texture_get_height(LwTexture *self)
{
	return self->priv->height;
}

/**
 * lw_texture_enable:
 * @self: A #LwTexture
 *
 * Enables the texture target and binds the texture name to the target.
 */
void
lw_texture_enable(LwTexture *self)
{
	glEnable(self->priv->target);
	glBindTexture(self->priv->target, self->priv->name);
}

/**
 * lw_texture_disable:
 * @self: A #LwTexture
 *
 * Disables the texture target.
 */
void
lw_texture_disable(LwTexture *self)
{
	glBindTexture(self->priv->target, 0);
	glDisable(self->priv->target);
}

static void
lw_texture_init(LwTexture *self)
{
	self->priv = LW_TEXTURE_GET_PRIVATE(self);

	self->priv->name = 0;
	self->priv->target = GL_TEXTURE_2D;
	self->priv->filter = GL_NEAREST;
	self->priv->wrap = GL_CLAMP_TO_EDGE;

	self->priv->width = 0;
	self->priv->height = 0;
	
	self->matrix = identity_texture_matrix;

	glGenTextures(1, &(self->priv->name));
}

static void
lw_texture_finalize(GObject *object)
{
	LwTexture *self = LW_TEXTURE(object);

	if(self->priv->name)
		glDeleteTextures(1, &(self->priv->name));

	/* Chain up to the parent class */
	G_OBJECT_CLASS(lw_texture_parent_class)->finalize(object);
}

static void
lw_texture_class_init(LwTextureClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = lw_texture_set_property;
	gobject_class->get_property = lw_texture_get_property;
	gobject_class->finalize = lw_texture_finalize;

	g_type_class_add_private(klass, sizeof(LwTexturePrivate));

	/**
	 * LwTexture:width:
	 *
	 * The width of the texture
	 */
	obj_properties[PROP_WIDTH] =
		g_param_spec_uint("width",
		                  "Width",
		                  "The width of the texture",
		                  0,
		                  G_MAXUINT,
		                  0,
		                  G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * LwTexture:height:
	 *
	 * The height of the texture
	 */
	obj_properties[PROP_HEIGHT] =
		g_param_spec_uint("height",
		                  "Height",
		                  "The height of the texture",
		                  0,
		                  G_MAXUINT,
		                  0,
		                  G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

