/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <stdio.h>
#include <livewallpaper/core.h>
#include <libpeas/peas.h>

#include "light.h"
#include "particle.h"
#include "galaxy.h"


#define FOVY 45.0f
#define NEAR  0.1f
#define FAR  20.0f

/*
 * Update this value, if FOVY or NEAR changes:
 * YMAX = NEAR * tanf (FOVY * _PI / 360.0)
 */
#define YMAX 0.04142135
#define DEG2RAD (_PI / 180.0f)

#define GALAXY_PLUGIN_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GALAXY_TYPE_PLUGIN, GalaxyPluginPrivate))

struct _GalaxyPluginPrivate
{
	gint rotate_x;
	gint rotate_y;
	gint rotate_z;

	double zoom;

	double offset_x;
	double offset_y;

	LwTexture *lightTexture;
	LwBackground *background;

	gboolean lp_enabled;
	GalaxyLightProgram *lp;

	GalaxyParticleSystem *ps;
};

enum
{
    PROP_0,

    PROP_CUSTOM_LIGHT,

    PROP_ROTATE_X,
    PROP_ROTATE_Y,
    PROP_ROTATE_Z,

    PROP_ZOOM,

    PROP_OFFSET_X,
    PROP_OFFSET_Y,

    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

static void lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface);

G_DEFINE_DYNAMIC_TYPE_EXTENDED(GalaxyPlugin, galaxy_plugin, PEAS_TYPE_EXTENSION_BASE, 0,
							   G_IMPLEMENT_INTERFACE_DYNAMIC(LW_TYPE_LIVEWALLPAPER_PLUGIN,
							   								 lw_livewallpaper_plugin_iface_init))


static void
galaxy_plugin_set_property(GObject *object,
                           guint property_id,
                           const GValue *value,
                           GParamSpec *pspec)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(object);

	switch(property_id)
	{
		case PROP_CUSTOM_LIGHT:
			if(GLEW_ARB_fragment_program)
				self->priv->lp_enabled = g_value_get_boolean(value);
			break;

		case PROP_ROTATE_X:
			self->priv->rotate_x = g_value_get_int(value);
			break;

		case PROP_ROTATE_Y:
			self->priv->rotate_y = g_value_get_int(value);
			break;

		case PROP_ROTATE_Z:
			self->priv->rotate_z = g_value_get_int(value);
			break;

		case PROP_ZOOM:
			self->priv->zoom = g_value_get_double(value);
			break;

		case PROP_OFFSET_X:
			self->priv->offset_x = g_value_get_double(value);
			break;

		case PROP_OFFSET_Y:
			self->priv->offset_y = g_value_get_double(value);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
galaxy_plugin_get_property(GObject *object,
                           guint property_id,
                           GValue *value,
                           GParamSpec *pspec)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(object);

	switch(property_id)
	{
		case PROP_CUSTOM_LIGHT:
			if(GLEW_ARB_fragment_program)
				g_value_set_boolean(value, self->priv->lp_enabled);
			break;

		case PROP_ROTATE_X:
			g_value_set_int(value, self->priv->rotate_x);
			break;

		case PROP_ROTATE_Y:
			g_value_set_int(value, self->priv->rotate_y);
			break;

		case PROP_ROTATE_Z:
			g_value_set_int(value, self->priv->rotate_z);
			break;

		case PROP_ZOOM:
			g_value_set_double(value, self->priv->zoom);
			break;

		case PROP_OFFSET_X:
			g_value_set_double(value, self->priv->offset_x);
			break;

		case PROP_OFFSET_Y:
			g_value_set_double(value, self->priv->offset_y);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
galaxy_plugin_adjust_viewport(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	double xmax = YMAX * lw_output_get_aspect_ratio(output);
	
	(void) plugin;

	/* Enable blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE); /* Additive */

	/* Save and reset the coordinate system before touching. */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	/*
	 * Using glFrustum instead of gluPrespective.
	 * http://www.opengl.org/wiki/GluPerspective_code
	 */
	glFrustum(-xmax, xmax, -YMAX, YMAX, NEAR, FAR);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
}

static void
galaxy_plugin_prepare_paint(LwLivewallpaperPlugin *plugin, int ms_since_last_paint)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(plugin);

	/* Update color values */
	if(self->priv->lp_enabled)
		galaxy_light_program_update(self->priv->lp);

	/* Update particles */
	galaxy_particle_system_update(self->priv->ps, ms_since_last_paint);
}

static void
galaxy_plugin_paint(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(plugin);

	/* Clear color buffer and draw background image */
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	lw_background_draw(self->priv->background, output);

	/* Apply the user defined transformations. */
	glLoadIdentity();

	/* Apply galaxy zoom. */
	glTranslatef(0.0f, 0.0f,
		-self->priv->zoom * lw_output_get_aspect_ratio(output));

	/* Apply galaxy offset and rotation. */
	glTranslatef(self->priv->offset_x, self->priv->offset_y, 0.0f);

	glRotatef(self->priv->rotate_x, 1.0f, 0.0f, 0.0f);
	glRotatef(self->priv->rotate_y, 0.0f, 1.0f, 0.0f);
	glRotatef(self->priv->rotate_z, 0.0f, 0.0f, 1.0f);

	/* Draw particles. */
	galaxy_particle_system_draw(self->priv->ps);

	/* Draw light. */
	if(!self->priv->lightTexture) return;

	lw_texture_enable(self->priv->lightTexture);

	/* Save texture environment. */
	glPushAttrib(GL_TEXTURE_BIT);

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	/* Enable light program */
	if(self->priv->lp_enabled)
		galaxy_light_program_enable(self->priv->lp);

	/* Draw quad. */
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
		glTexCoord2d(1.0f, 0.0f);
		glVertex2f(-1.0f, -1.0f);

		glTexCoord2d(0.0f, 0.0f);
		glVertex2f(1.0f, -1.0f);

		glTexCoord2d(0.0f, 1.0f);
		glVertex2f(1.0f, 1.0f);

		glTexCoord2d(1.0f, 1.0f);
		glVertex2f(-1.0f, 1.0f);
	glEnd();

	/* Disable light program */
	if(self->priv->lp_enabled)
		galaxy_light_program_disable(self->priv->lp);

	/* Restore texture environment. */
	glPopAttrib();

	lw_texture_disable(self->priv->lightTexture);
}

static void
galaxy_plugin_restore_viewport(LwLivewallpaperPlugin *plugin)
{
	(void) plugin;

	/* Restore the coordinate system. */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	/* Restore the original perspective. */
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	/* Restore blending and other stuffs. */
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
}

static void
galaxy_plugin_init(GalaxyPlugin *self)
{
	self->priv = GALAXY_PLUGIN_GET_PRIVATE(self);

	self->priv->rotate_x = -60;
	self->priv->rotate_y = 15;
	self->priv->rotate_z = 0;

	self->priv->zoom = 1.5f;

	self->priv->offset_x = 0.0f;
	self->priv->offset_y = 0.0f;

	self->priv->lp_enabled = FALSE;
	self->priv->lp = g_object_new(GALAXY_TYPE_LIGHT_PROGRAM, NULL);

	self->priv->ps = g_object_new(GALAXY_TYPE_PARTICLE_SYSTEM, NULL);
}

static void
galaxy_plugin_init_plugin(LwLivewallpaperPlugin *plugin)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(plugin);
	gchar *filename, *datadir = peas_extension_base_get_data_dir(PEAS_EXTENSION_BASE(self));
	GSettings *settings = g_settings_new("net.launchpad.livewallpaper.plugins.galaxy");
	GError *error = NULL;

	/* Load light texture. */
	filename = g_build_filename(datadir, "galaxy-light.png", NULL);
	self->priv->lightTexture = lw_texture_new_from_file(filename, &error);
	if(error != NULL)
	{
		g_warning("Could not load light texture: %s", error->message);
		g_error_free(error);
		error = NULL;
	}
	g_free(filename);

	/* Load space texture as background. */
	filename = g_build_filename(datadir, "space.png", NULL);
	self->priv->background = lw_background_new_with_default_image(filename, LwBackgroundTiled, &error);
	if(error != NULL)
	{
		g_warning("Could not load space texture: %s", error->message);
		g_error_free(error);
		error = NULL;
	}
	g_free(filename);

	/* Show message if the custom galaxy light is not supported */
	if(!GLEW_ARB_fragment_program)
		g_message("Custom galaxy light is not supported.");

	/* Bind settings to properties. */
	g_settings_bind(settings, "rotate-x", self, "rotate-x", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "rotate-y", self, "rotate-y", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "rotate-z", self, "rotate-z", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "zoom", self, "zoom", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "offset-x", self, "offset-x", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "offset-y", self, "offset-y", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "enable-custom-light", self, "enable-custom-light", G_SETTINGS_BIND_GET);
	g_settings_bind_with_mapping(settings, "outer-light-color", self->priv->lp, "outer-color", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping,NULL, NULL);
	g_settings_bind_with_mapping(settings, "inner-light-color", self->priv->lp, "inner-color", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping,NULL, NULL);
	g_settings_bind(settings, "color-radius", self->priv->lp, "color-radius", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "star-count", self->priv->ps, "star-count", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "star-size", self->priv->ps, "star-size", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "draw-streaks", self->priv->ps, "draw-streaks", G_SETTINGS_BIND_GET);
}

static void
galaxy_plugin_dispose(GObject *object)
{
	GalaxyPlugin *self = GALAXY_PLUGIN(object);

	if(self->priv->lightTexture)
	{
		g_object_unref(self->priv->lightTexture);
		self->priv->lightTexture = NULL;
	}

	if(self->priv->background)
	{
		g_object_unref(self->priv->background);
		self->priv->background = NULL;
	}

	if(self->priv->lp)
	{
		g_object_unref(self->priv->lp);
		self->priv->lp = NULL;
	}

	if(self->priv->ps)
	{
		g_object_unref(self->priv->ps);
		self->priv->ps = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(galaxy_plugin_parent_class)->dispose(object);
}

static void
galaxy_plugin_class_init(GalaxyPluginClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = galaxy_plugin_set_property;
	gobject_class->get_property = galaxy_plugin_get_property;
	gobject_class->dispose = galaxy_plugin_dispose;

	g_type_class_add_private(klass, sizeof(GalaxyPluginPrivate));

	obj_properties[PROP_CUSTOM_LIGHT] =
		g_param_spec_boolean("enable-custom-light",
		                     "Enable custom light color",
		                     "Enable custom light color",
		                     FALSE,
		                     G_PARAM_READWRITE);

	obj_properties[PROP_ROTATE_X] =
		g_param_spec_int("rotate-x",
		                 "Rotate X",
		                 "Rotate the galaxy through X coordinate",
		                 -90,
		                 90,
		                 -60,
		                 G_PARAM_READWRITE);

	obj_properties[PROP_ROTATE_Y] =
		g_param_spec_int("rotate-y",
		                 "Rotate Y",
		                 "Rotate the galaxy through Y coordinate",
		                 -90,
		                 90,
		                 15,
		                 G_PARAM_READWRITE);

	obj_properties[PROP_ROTATE_Z] =
		g_param_spec_int("rotate-z",
		                 "Rotate Z",
		                 "Rotate the galaxy through Z coordinate",
		                 0,
		                 360,
		                 0,
		                 G_PARAM_READWRITE);

	obj_properties[PROP_ZOOM] =
		g_param_spec_double("zoom",
		                    "Zoom",
		                    "Zoom in and out",
		                    0.0f,
		                    5.0f,
		                    1.5f,
		                    G_PARAM_READWRITE);

	obj_properties[PROP_OFFSET_X] =
		g_param_spec_double("offset-x",
		                    "Offset X",
		                    "X offset of the centre of the galaxy",
		                    -2.0f,
		                    2.0f,
		                    0.0f,
		                    G_PARAM_READWRITE);

	obj_properties[PROP_OFFSET_Y] =
		g_param_spec_double("offset-y",
		                    "Offset Y",
		                    "Y offset of the centre of the galaxy",
		                    -2.0f,
		                    2.0f,
		                    0.0f,
		                    G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

static void
galaxy_plugin_class_finalize(GalaxyPluginClass *klass)
{
	(void) klass;
}

static void
lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface)
{
	iface->init_plugin = galaxy_plugin_init_plugin;

	iface->adjust_viewport = galaxy_plugin_adjust_viewport;
	iface->prepare_paint = galaxy_plugin_prepare_paint;
	iface->paint = galaxy_plugin_paint;
	iface->restore_viewport = galaxy_plugin_restore_viewport;
}

G_MODULE_EXPORT void
peas_register_types(PeasObjectModule *module)
{
	galaxy_plugin_register_type(G_TYPE_MODULE(module));
	peas_object_module_register_extension_type(module, LW_TYPE_LIVEWALLPAPER_PLUGIN, GALAXY_TYPE_PLUGIN);
}

