/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <string.h>
#include <glib-object.h>
#include <livewallpaper/core.h>
#include "light.h"

/*
 * The light program calculates the resulting color like GIMP's Color layer mode.
 * This means we use the hue and saturation of the specified color and the lightness
 * of the texture. To get a small ARB fragment program we precalculate some parts of
 * the color (HSL to RGB conversion is to complex for ARB, see preparePaint). The
 * fragment program only combines the precalculated values with the lightness value.
 */
static const char source[] =
	"!!ARBfp1.0\n"
	"TEMP lightness;\n"
	"TEMP texcolor;\n"

	/* Outer light color. */
	"TEMP color0;\n"
	"TEMP color0_p, color0_v;\n"

	/* Inner light color. */
	"TEMP color1;\n"
	"TEMP color1_p, color1_v;\n"

	/* Used for radial color gradient. */
	"TEMP radius;\n"

	/* Used to simulate if statement. */
	"TEMP if;\n"

	/* Get lightness from texture. */
	"TEX texcolor, fragment.texcoord[0], texture[0], 2D;\n"

	"MAX lightness.xy, texcolor.xyxy, texcolor.yzyz;\n"
	"MAX lightness.x, lightness.x, lightness.y;\n"
	"MIN lightness.zw, texcolor.xyxy, texcolor.yzyz;\n"
	"MIN lightness.y, lightness.z, lightness.w;\n"

	"ADD lightness.x, lightness.x, lightness.y;\n"
	"MUL lightness.x, lightness.x, {0.5, 0.5, 0.5, 0.5};\n"

	/* Case 1: lightness <= 0.5f */
	"SGE if, {0.5, 0.5, 0.5, 0.0}, lightness.x;\n"
	"MUL color0_p, if, program.local[0];\n"
	"MUL color0_v, if, program.local[1];\n"

	"MUL color1_p, if, program.local[4];\n"
	"MUL color1_v, if, program.local[5];\n"

	/* Case 2: lightness > 0.5f */
	"SLT if, {0.5, 0.5, 0.5, 0.0}, lightness.x;\n"
	"MAD color0_p, if, program.local[2], color0_p;\n"
	"MAD color0_v, if, program.local[3], color0_v;\n"

	"MAD color1_p, if, program.local[6], color1_p;\n"
	"MAD color1_v, if, program.local[7], color1_v;\n"

	/* Calculate color. */
	"MAD color0, color0_v, lightness.x, color0_p;\n"

	"MAD color1, color1_v, lightness.x, color1_p;\n"

	/* Calculate distance to center. */
	"SUB radius, fragment.texcoord[0], {0.5, 0.5, 0.5, 0.5};\n"
	"MUL radius, radius, radius;\n"
	"ADD radius.x, radius.x, radius.y;\n"
	"RSQ radius, radius.x;\n"
	"RCP radius, radius.x;\n"
	"MUL radius, radius, program.local[8].x;\n"

	/* Set result. */
	"LRP result.color, radius, color0, color1;\n"
	"END";

#define GALAXY_LIGHT_PROGRAM_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GALAXY_TYPE_LIGHT_PROGRAM, GalaxyLightProgramPrivate))

struct _GalaxyLightProgramPrivate
{
	GLuint name;

	GdkRGBA outer_color;
	GdkRGBA inner_color;

	double color_radius;
};

enum
{
    PROP_0,

	PROP_OUTER_COLOR,
	PROP_INNER_COLOR,

	PROP_COLOR_RADIUS,

    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(GalaxyLightProgram, galaxy_light_program, G_TYPE_OBJECT)


static void
galaxy_light_program_set_property(GObject *object,
                                    guint property_id,
                                    const GValue *value,
                                    GParamSpec *pspec)
{
	GalaxyLightProgram *self = GALAXY_LIGHT_PROGRAM(object);

	switch(property_id)
	{
		case PROP_OUTER_COLOR:
			galaxy_light_program_set_outer_color(self, g_value_get_boxed(value));
			break;

		case PROP_INNER_COLOR:
			galaxy_light_program_set_inner_color(self, g_value_get_boxed(value));
			break;

		case PROP_COLOR_RADIUS:
			galaxy_light_program_set_color_radius(self, g_value_get_double(value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
galaxy_light_program_get_property(GObject *object,
                                    guint property_id,
                                    GValue *value,
                                    GParamSpec *pspec)
{
	GalaxyLightProgram *self = GALAXY_LIGHT_PROGRAM(object);

	switch(property_id)
	{
		case PROP_OUTER_COLOR:
		{
			GdkRGBA rgba;
			galaxy_light_program_get_outer_color(self, &rgba);
			g_value_set_boxed(value, &rgba);
			break;
		}

		case PROP_INNER_COLOR:
		{
			GdkRGBA rgba;
			galaxy_light_program_get_inner_color(self, &rgba);
			g_value_set_boxed(value, &rgba);
			break;
		}

		case PROP_COLOR_RADIUS:
			g_value_set_double(value, galaxy_light_program_get_color_radius(self));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static inline float
hsl_value(float n1, float n2, float hue)
{
	if(hue < 0.0f)
		hue += 6.0f;
	else if(hue > 6.0f)
		hue -= 6.0f;

	if(hue < 1.0f)
		return n1 + (n2 - n1) * hue;
	else if(hue < 3.0f)
		return n2;
	else if(hue < 4.0f)
		return n1 + (n2 - n1) * (4.0f - hue);
	else
		return n1;
}

#define set_hsl_field(field, n1, n2, hue)            \
	field[0] = hsl_value(n1, n2, hue * 6.0f + 2.0f); \
	field[1] = hsl_value(n1, n2, hue * 6.0f);        \
	field[2] = hsl_value(n1, n2, hue * 6.0f - 2.0f)

static void
galaxy_light_program_set_arb_params(gint index,
                                    float r,
                                    float g,
                                    float b)
{
	float p1, p2, v1, v2;
	float f[3];

	/* Convert RGB to HSL */
	lw_rgb_to_hsl(&r, &g, &b);

#define h r
#define s g
#define l b

	/*
	 * We won't convert HSL to RGB in ARB so we calculate some helping values p and v
	 * here. In ARB we calculate the color in the following way:
	 *     color = p + v * lightness
	 * 
	 * We convert it in the same way like the hsl_to_rgb function we only leave the
	 * lightness and add it later (We don't know the lightness at the moment).
	 */

	/* Case 1: lightness <= 0.5f */
		p1 = p2 = 0.0f;

		v2 = 1.0f + s;
		v1 = 2.0f - v2;

		set_hsl_field(f, p1, p2, h);
		glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, index, f[0], f[1], f[2], 1.0f);

		set_hsl_field(f, v1, v2, h);
		glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, index + 1, f[0], f[1], f[2], 0.0f);

	/* Case 2: lightness > 0.5f */
		p2 = s;
		p1 = -p2;

		v2 = 1.0f - s;
		v1 = 2.0f - v2;

		set_hsl_field(f, p1, p2, h);
		glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, index + 2, f[0], f[1], f[2], 1.0f);

		set_hsl_field(f, v1, v2, h);
		glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, index + 3, f[0], f[1], f[2], 0.0f);

#undef h
#undef s
#undef l
}

#undef set_hsl_field

void
galaxy_light_program_update(GalaxyLightProgram *self)
{
	if(!GLEW_ARB_fragment_program) return;

	/* Bind light program to set local parameters */
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, self->priv->name);

	/* Set outer light color */
	galaxy_light_program_set_arb_params(0, self->priv->outer_color.red,
	                                       self->priv->outer_color.green,
	                                       self->priv->outer_color.blue);

	/* Set inner light color */
	galaxy_light_program_set_arb_params(4, self->priv->inner_color.red,
	                                       self->priv->inner_color.green,
	                                       self->priv->inner_color.blue);

	/* Set color radius */
	glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 8,
		2.0f * self->priv->color_radius, 0.0f, 0.0f, 0.0f);
}

void
galaxy_light_program_set_outer_color(GalaxyLightProgram *self, GdkRGBA *color)
{
	self->priv->outer_color = *color;
}

void
galaxy_light_program_get_outer_color(GalaxyLightProgram *self, GdkRGBA *color)
{
	*color = self->priv->outer_color;
}

void
galaxy_light_program_set_inner_color(GalaxyLightProgram *self, GdkRGBA *color)
{
	self->priv->inner_color = *color;
}

void
galaxy_light_program_get_inner_color(GalaxyLightProgram *self, GdkRGBA *color)
{
	*color = self->priv->inner_color;
}

void
galaxy_light_program_set_color_radius(GalaxyLightProgram *self, double radius)
{
	self->priv->color_radius = radius;
}

double
galaxy_light_program_get_color_radius(GalaxyLightProgram *self)
{
	return self->priv->color_radius;
}

void
galaxy_light_program_enable(GalaxyLightProgram *self)
{
	if(!GLEW_ARB_fragment_program) return;

	glEnable(GL_FRAGMENT_PROGRAM_ARB);
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, self->priv->name);
}

void
galaxy_light_program_disable(GalaxyLightProgram *self)
{
	(void) self;

	if(!GLEW_ARB_fragment_program) return;

	glDisable(GL_FRAGMENT_PROGRAM_ARB);
}

static void
galaxy_light_program_init(GalaxyLightProgram *self)
{
    self->priv = GALAXY_LIGHT_PROGRAM_GET_PRIVATE(self);

    self->priv->name = 0;
    if(GLEW_ARB_fragment_program)
    {
		/* Clear errors */
		glGetError();

		/* Create program */
		glGenProgramsARB(1, &self->priv->name);

		glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, self->priv->name);
		glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB,
			strlen(source), source);

		/* Handle errors */
		if(glGetError() == GL_INVALID_OPERATION)
		{
			const char *errorStr;

			/* GLint errorPos;
			glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &errorPos); */

			errorStr = (const char *) glGetString(GL_PROGRAM_ERROR_STRING_ARB);
			g_critical("Failed to load light program: %s", errorStr);

			glDeleteProgramsARB(1, &self->priv->name);
			self->priv->name = 0;
		}
    }
}

static void
galaxy_light_program_finalize(GObject *object)
{
	GalaxyLightProgram *self = GALAXY_LIGHT_PROGRAM(object);

	if(self->priv->name != 0 && GLEW_ARB_fragment_program)
		glDeleteProgramsARB(1, &self->priv->name);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(galaxy_light_program_parent_class)->finalize(object);
}

static void
galaxy_light_program_class_init(GalaxyLightProgramClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = galaxy_light_program_set_property;
	gobject_class->get_property = galaxy_light_program_get_property;
	gobject_class->finalize = galaxy_light_program_finalize;

	g_type_class_add_private(klass, sizeof(GalaxyLightProgramPrivate));

	obj_properties[PROP_OUTER_COLOR] =
		g_param_spec_boxed("outer-color",
		                   "Outer Color",
		                   "Color of the outer parts of the galaxy",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	obj_properties[PROP_INNER_COLOR] =
		g_param_spec_boxed("inner-color",
		                   "Inner Color",
		                   "Color in the centre of the galaxy",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	obj_properties[PROP_COLOR_RADIUS] =
		g_param_spec_double("color-radius",
		                    "Color radius",
		                    "Radius of the radial color gradient",
		                    0.0,
		                    2.0,
		                    1.0,
		                    G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

