/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <glib-object.h>
#include <livewallpaper/core.h>
#include "particle.h"


typedef struct _Star Star;

struct _Star
{
	/*
	 * Those values needed for the general parametic form of an
	 * ellipse. Phi is the angle between the x-axis and the major
	 * axis of the ellipse.
	 */
	float cos_phi, sin_phi;

	/* The current angle/position of this star. */
	float angle;

	/* The z value of the real position of this star. */
	float z;

	/* The distance from the center of the galaxy to this star. */
	float distance;

	/* the angular velocity of the star in 1/ms. */
	float speed;
};

#define STAR_COLOR_R 255
#define STAR_COLOR_G 244
#define STAR_COLOR_B 234
#define STAR_COLOR_A 200

#define _PI  (float) 3.141593
#define _2PI (float) (2 * _PI)

/* 0 (line) -> 1.0f (circle) */
#define ELLIPSE_RATIO .885f

#define GALAXY_PARTICLE_SYSTEM_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GALAXY_TYPE_PARTICLE_SYSTEM, GalaxyParticleSystemPrivate))

struct _GalaxyParticleSystemPrivate
{
	guint star_count;
	guint star_size;
	gboolean draw_streaks;

	GArray *stars;
	float *vertices;

	LwTexture *flareTexture;
};

enum
{
    PROP_0,

    PROP_STAR_COUNT,
    PROP_STAR_SIZE,
    PROP_DRAW_STREAKS,

    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(GalaxyParticleSystem, galaxy_particle_system, G_TYPE_OBJECT)


static void
galaxy_particle_system_set_property(GObject *object,
                                    guint property_id,
                                    const GValue *value,
                                    GParamSpec *pspec)
{
	GalaxyParticleSystem *self = GALAXY_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_STAR_COUNT:
			galaxy_particle_system_set_star_count(self, g_value_get_uint(value));
			break;

		case PROP_STAR_SIZE:
			galaxy_particle_system_set_star_size(self, g_value_get_uint(value));
			break;
		
		case PROP_DRAW_STREAKS:
			galaxy_particle_system_set_draw_streaks(self, g_value_get_boolean(value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
galaxy_particle_system_get_property(GObject *object,
                                    guint property_id,
                                    GValue *value,
                                    GParamSpec *pspec)
{
	GalaxyParticleSystem *self = GALAXY_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_STAR_COUNT:
			g_value_set_uint(value, galaxy_particle_system_get_star_count(self));
			break;

		case PROP_STAR_SIZE:
			g_value_set_uint(value, galaxy_particle_system_get_star_size(self));
			break;

		case PROP_DRAW_STREAKS:
			g_value_set_boolean(value, galaxy_particle_system_get_draw_streaks(self));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

/*
 * Gaussian random generator
 * Using the polar form of the Box-Muller transformation
 */
static float
randng(void)
{
	static float y1, y2;
	static gboolean have_next = FALSE;

	if (! have_next)
	{
		float x1, x2, w;
		
		do
		{
			x1 = 2.0 * randf() - 1.0;
			x2 = 2.0 * randf() - 1.0;
			w = x1 * x1 + x2 * x2;
		} while(w >= 1.0);

		w = sqrt(-2.0 * log (w) / w);
		y1 = x1 * w;
		y2 = x2 * w;
		have_next = TRUE;

		return y1;
	}
	/* else */
	have_next = FALSE;
	return y2;
}

static inline void
galaxy_particle_system_init_star(Star *star)
{
	star->angle = rand1f(_2PI);

	star->distance = fabs(randng()) * 0.5f;

	/*
	 * Phi depends on the distance of the star to get density waves.
	 * See http://en.wikipedia.org/wiki/Density_wave_theory
	 */
	star->cos_phi = cos(star->distance * _2PI);
	star->sin_phi = sin(star->distance * _2PI);

	star->z = (1 - star->distance) * 0.25f * randng()
	          / (_PI + _2PI * star->distance * star->distance);

	if(star->distance < 0.2f)
		star->z *= 0.8f;

	/* The angular velocity of the star in 1/ms */
	star->speed = -rand2f(0.00003f, 0.000045f) / star->distance;
}

void
galaxy_particle_system_set_star_count(GalaxyParticleSystem *self, guint count)
{
	g_array_set_size(self->priv->stars, count);

	/* Append stars */
	if(self->priv->star_count < count)
	{
		guint i;

		for(i = self->priv->star_count; i < count; i++)
		{
			Star *star = &g_array_index(self->priv->stars, Star, i);
			galaxy_particle_system_init_star(star);
		}
	}

	self->priv->vertices = g_realloc(self->priv->vertices, 3 * count * sizeof(float));
	self->priv->star_count = count;
}

guint
galaxy_particle_system_get_star_count(GalaxyParticleSystem *self)
{
	return self->priv->star_count;
}

void
galaxy_particle_system_set_star_size(GalaxyParticleSystem *self, guint size)
{
	self->priv->star_size = size;
}

guint
galaxy_particle_system_get_star_size(GalaxyParticleSystem *self)
{
	return self->priv->star_size;
}

#define FLARE_SIZE 32
static void
galaxy_particle_system_generate_flare(GalaxyParticleSystem *self)
{
	guchar *data = (guchar*) g_malloc(FLARE_SIZE * FLARE_SIZE * 4);
	gint x;

	for(x = 0; x < FLARE_SIZE; x++)
	{
		gint y;
		for(y = 0; y < FLARE_SIZE; y++)
		{
			float dx = 1.0f - 2.0f * x / FLARE_SIZE,
				  dy = 1.0f - 2.0f * y / FLARE_SIZE,
				  r  = 1.0f - sqrt(dx * dx + dy * dy),
				  /* Draw flare */
				  fa = (r > 1) ? 1.0f : r*r;	
			gint i = (x + y * FLARE_SIZE) * 4;

			/* Draw streaks */
			if(self->priv->draw_streaks && ((dx == 0.0f) || (dy == 0.0f)))
				fa += r * 0.5;

			if(fa > 1)
				fa = 1.0f;

			data[i + 0] = 255; /* Red */
			data[i + 1] = 255; /* Green */
			data[i + 2] = 255; /* Blue */
			data[i + 3] = (guchar) (255 * fa); /* Alpha */
		}
	}

	self->priv->flareTexture = lw_texture_new_from_data(data, FLARE_SIZE, FLARE_SIZE, GL_RGBA, GL_UNSIGNED_BYTE);
	g_free(data);

	if(self->priv->flareTexture)
	{
		lw_texture_set_filter(self->priv->flareTexture, GL_LINEAR);
		lw_texture_set_wrap(self->priv->flareTexture, GL_CLAMP_TO_EDGE);
	}
}
#undef FLARE_SIZE

void
galaxy_particle_system_set_draw_streaks(GalaxyParticleSystem *self, gboolean draw_streaks)
{
	if(self->priv->draw_streaks == draw_streaks)
		return;

	self->priv->draw_streaks = draw_streaks;
	galaxy_particle_system_generate_flare(self);
}

gboolean
galaxy_particle_system_get_draw_streaks(GalaxyParticleSystem *self)
{
	return self->priv->draw_streaks;
}

void
galaxy_particle_system_update(GalaxyParticleSystem *self, int ms_since_last_paint)
{
	float *v = self->priv->vertices;
	guint i;

	for(i = 0; i < self->priv->star_count; i++)
	{
		float a_cos_angle, b_sin_angle;
		Star *star = &g_array_index(self->priv->stars, Star, i);

		star->angle += ms_since_last_paint * star->speed;

		/*
		 * a is the semi-major axis, here the distance and b is the
		 * semi-minor axis defined by the distance and the ellipse ratio.
		 */
		a_cos_angle = star->distance * cos(star->angle);
		b_sin_angle = star->distance * ELLIPSE_RATIO * sin(star->angle);

		/*
		 * To calculate the current position we use the general
		 * parametic form of an ellipse.
		 * See http://en.wikipedia.org/wiki/Ellipse#General_parametric_form
		 */
		v[0] = a_cos_angle * star->cos_phi - b_sin_angle * star->sin_phi;
		v[1] = a_cos_angle * star->sin_phi + b_sin_angle * star->cos_phi;
		v[2] = star->z;
		v += 3;
	}
}

void
galaxy_particle_system_draw(GalaxyParticleSystem *self)
{
	static GLfloat quadratic[] = {0.4f, 0.0f, 0.1f};
	
	if(self->priv->flareTexture) lw_texture_enable(self->priv->flareTexture);

	/* Save texture environment */
	glPushAttrib(GL_TEXTURE_BIT);

	glPointParameterfv(GL_POINT_DISTANCE_ATTENUATION, quadratic);
	glPointParameterf(GL_POINT_FADE_THRESHOLD_SIZE, 60.0f);
	glPointParameterf(GL_POINT_SIZE_MIN, 0.1f);

	glPointSize(self->priv->star_size);

	glTexEnvi(GL_POINT_SPRITE_ARB, GL_COORD_REPLACE, GL_TRUE);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* Render stars using vertex array */
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnable(GL_POINT_SPRITE);
	glColor4ub(STAR_COLOR_R, STAR_COLOR_G, STAR_COLOR_B, STAR_COLOR_A);

	glVertexPointer(3, GL_FLOAT, 3 * sizeof(float), self->priv->vertices);
	glDrawArrays(GL_POINTS, 0, self->priv->star_count);

	glDisable(GL_POINT_SPRITE);
	glDisableClientState(GL_VERTEX_ARRAY);

	/* Restore texture environment */
	glPopAttrib();

	if(self->priv->flareTexture) lw_texture_disable(self->priv->flareTexture);
}

static void
galaxy_particle_system_init(GalaxyParticleSystem *self)
{
    self->priv = GALAXY_PARTICLE_SYSTEM_GET_PRIVATE(self);

    self->priv->star_count = 0;
    self->priv->star_size = 8;
    self->priv->draw_streaks = TRUE;
    galaxy_particle_system_generate_flare(self);

    self->priv->stars = g_array_new(FALSE, TRUE, sizeof(Star));
    self->priv->vertices = NULL;
}

static void
galaxy_particle_system_dispose(GObject *object)
{
	GalaxyParticleSystem *self = GALAXY_PARTICLE_SYSTEM(object);

	if(self->priv->flareTexture)
	{
		g_object_unref(self->priv->flareTexture);
		self->priv->flareTexture = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(galaxy_particle_system_parent_class)->dispose(object);
}

static void
galaxy_particle_system_finalize(GObject *object)
{
	GalaxyParticleSystem *self = GALAXY_PARTICLE_SYSTEM(object);

	g_array_free(self->priv->stars, TRUE);
	g_free(self->priv->vertices);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(galaxy_particle_system_parent_class)->finalize(object);
}

static void
galaxy_particle_system_class_init(GalaxyParticleSystemClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = galaxy_particle_system_set_property;
	gobject_class->get_property = galaxy_particle_system_get_property;
	gobject_class->dispose = galaxy_particle_system_dispose;
	gobject_class->finalize = galaxy_particle_system_finalize;

	g_type_class_add_private(klass, sizeof(GalaxyParticleSystemPrivate));

	obj_properties[PROP_STAR_COUNT] =
		g_param_spec_uint("star-count",
		                  "Number of stars",
		                  "Number of stars",
		                  0,
		                  500000,
		                  0,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_STAR_SIZE] =
		g_param_spec_uint("star-size",
		                  "Star size",
		                  "Size of a star",
		                  0,
		                  64,
		                  8,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_DRAW_STREAKS] =
		g_param_spec_boolean("draw-streaks",
		                     "Draw star streaks",
		                     "Draw a star with streaks",
		                     TRUE,
		                     G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

