/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <gdk/gdk.h>

#include <livewallpaper/core.h>
#include <libpeas/peas.h>

#include "particle.h"
#include "nexus.h"


#define NEXUS_PLUGIN_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NEXUS_TYPE_PLUGIN, NexusPluginPrivate))

struct _NexusPluginPrivate
{
	NexusParticleSystem *ps;
	
	LwBackground *background;
};

static void lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface);

G_DEFINE_DYNAMIC_TYPE_EXTENDED(NexusPlugin, nexus_plugin, PEAS_TYPE_EXTENSION_BASE, 0,
							   G_IMPLEMENT_INTERFACE_DYNAMIC(LW_TYPE_LIVEWALLPAPER_PLUGIN,
							   								 lw_livewallpaper_plugin_iface_init))


static void
nexus_plugin_adjust_viewport(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	guint width, height;
	(void) plugin;

	/* Enable blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	/* Save and change texture environment */
	glPushAttrib(GL_TEXTURE_BIT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* Save and reset the coordinate system before touching */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	width = lw_output_get_width(output);
	height = lw_output_get_height(output);

	if(width > height)
		glOrtho(0.0, 1.0, 0.0, ((double) height) / ((double) width), 0.0, 1.0);
	else
		glOrtho(0.0, ((double) width) / ((double) height), 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
}

static void
nexus_plugin_prepare_paint(LwLivewallpaperPlugin *plugin, int ms_since_last_paint)
{
	NexusPlugin *self = NEXUS_PLUGIN(plugin);

	nexus_particle_system_update(self->priv->ps, ms_since_last_paint);
}

static void
nexus_plugin_paint(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	NexusPlugin *self = NEXUS_PLUGIN(plugin);

	/* Clear color buffer and draw background image */
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	lw_background_draw(self->priv->background, output);

	/* Draw pulses */
	nexus_particle_system_draw(self->priv->ps, lw_output_get_max_side(output));
}

static void
nexus_plugin_restore_viewport(LwLivewallpaperPlugin *plugin)
{
	(void) plugin;

	/* Restore the coordinate system */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	/* Restore the original perspective */
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	/* Restore texture environment */
	glPopAttrib();

	/* Restore blending */
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
}

static void
nexus_plugin_update_glow_type(GSettings *settings, gchar *key, NexusPlugin *self)
{
	(void) key;

	nexus_particle_system_set_glow_type(self->priv->ps, g_settings_get_enum(settings, "glow-type"));
}

static void
nexus_plugin_init(NexusPlugin *self)
{
	self->priv = NEXUS_PLUGIN_GET_PRIVATE(self);

	self->priv->ps = g_object_new(NEXUS_TYPE_PARTICLE_SYSTEM, NULL);
}

static void
nexus_plugin_init_plugin(LwLivewallpaperPlugin *plugin)
{
	NexusPlugin *self = NEXUS_PLUGIN(plugin);
	GSettings *settings = g_settings_new("net.launchpad.livewallpaper.plugins.nexus");
	gchar *filename, *datadir = peas_extension_base_get_data_dir(PEAS_EXTENSION_BASE(self));
	GError *error = NULL;

	nexus_particle_system_set_texture_dir(self->priv->ps, datadir);

	/* Load background image */
	filename = g_build_filename(datadir, "background.png", NULL);
	self->priv->background = lw_background_new_with_default_image(filename, LwBackgroundZoom, &error);
	if(error != NULL)
	{
		g_warning("Could not load default background texture: %s", error->message);
		g_error_free(error);
		error = NULL;
	}
	g_free(filename);

	/* Bind settings to properties */
	g_settings_bind(settings, "pulse-count", self->priv->ps, "pulse-count", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "pulse-size", self->priv->ps, "pulse-size", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "max-length", self->priv->ps, "max-length", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "min-length", self->priv->ps, "min-length", G_SETTINGS_BIND_GET);

	g_settings_bind(settings, "max-delay", self->priv->ps, "max-delay", G_SETTINGS_BIND_GET);

	/* TODO: find a better way to bind enum to gint */
	g_signal_connect(settings, "changed::glow-type", G_CALLBACK(nexus_plugin_update_glow_type), self);
	nexus_particle_system_set_glow_type(self->priv->ps, g_settings_get_enum(settings, "glow-type"));
	/* g_settings_bind(settings, "glow-type", self->priv->ps, "glow-type", G_SETTINGS_BIND_GET); */

	g_settings_bind_with_mapping(settings, "color1", self->priv->ps, "color1", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
	g_settings_bind_with_mapping(settings, "color2", self->priv->ps, "color2", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
	g_settings_bind_with_mapping(settings, "color3", self->priv->ps, "color3", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
	g_settings_bind_with_mapping(settings, "color4", self->priv->ps, "color4", G_SETTINGS_BIND_GET,
	                             lw_color_get_color_mapping, lw_color_set_color_mapping, NULL, NULL);
}

static void
nexus_plugin_dispose(GObject *object)
{
	NexusPlugin *self = NEXUS_PLUGIN(object);

	if(self->priv->ps)
	{
		g_object_unref(self->priv->ps);
		self->priv->ps = NULL;
	}
	
	if(self->priv->background)
	{
		g_object_unref(self->priv->background);
		self->priv->background = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(nexus_plugin_parent_class)->dispose(object);
}

static void
nexus_plugin_class_init(NexusPluginClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = nexus_plugin_dispose;

	g_type_class_add_private(klass, sizeof(NexusPluginPrivate));
}

static void
nexus_plugin_class_finalize(NexusPluginClass *klass)
{
	(void) klass;
}

static void
lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface)
{
	iface->init_plugin = nexus_plugin_init_plugin;

	iface->adjust_viewport = nexus_plugin_adjust_viewport;
	iface->prepare_paint = nexus_plugin_prepare_paint;
	iface->paint = nexus_plugin_paint;
	iface->restore_viewport = nexus_plugin_restore_viewport;
}

G_MODULE_EXPORT void
peas_register_types(PeasObjectModule *module)
{
	nexus_plugin_register_type(G_TYPE_MODULE(module));
	peas_object_module_register_extension_type(module, LW_TYPE_LIVEWALLPAPER_PLUGIN, NEXUS_TYPE_PLUGIN);
}

