/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include <glib-object.h>
#include <gdk/gdk.h>
#include <livewallpaper/core.h>

#include "particle.h"


typedef struct _Pulse Pulse;

struct _Pulse
{
	/* Position of the pulse. */
	float x, y;

	/* Velocity of the pulse. */
	float vx, vy;

	/* Length of the pulse. */
	float length;

	/* Index of the pulse color. */
	int color;

	/* Pulse appears after delay milliseconds on the wallpaper. */
	int delay;
};

#define NEXUS_PARTICLE_SYSTEM_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NEXUS_TYPE_PARTICLE_SYSTEM, NexusParticleSystemPrivate))

struct _NexusParticleSystemPrivate
{
	guint pulse_count;
	guint pulse_size;

	double max_length;
	double min_length;

	guint max_delay;

	GArray *pulses;

	gchar *texturedir;
	guint glow_type;
	LwTexture *glowTexture;
	LwTexture *trailTexture;

	GdkRGBA colors[4];
};

enum
{
    PROP_0,

    PROP_PULSE_COUNT,
    PROP_PULSE_SIZE,

    PROP_MAX_LENGTH,
    PROP_MIN_LENGTH,

    PROP_MAX_DELAY,

    PROP_GLOW_TYPE,

    PROP_COLOR_1,
    PROP_COLOR_2,
    PROP_COLOR_3,
    PROP_COLOR_4,

    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(NexusParticleSystem, nexus_particle_system, G_TYPE_OBJECT)


static void
nexus_particle_system_set_property(GObject *object,
                                   guint property_id,
                                   const GValue *value,
                                   GParamSpec *pspec)
{
	NexusParticleSystem *self = NEXUS_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_PULSE_COUNT:
			nexus_particle_system_set_pulse_count(self, g_value_get_uint(value));
			break;

		case PROP_PULSE_SIZE:
			nexus_particle_system_set_pulse_size(self, g_value_get_uint(value));
			break;

		case PROP_MAX_LENGTH:
			nexus_particle_system_set_max_length(self, g_value_get_double(value));
			break;

		case PROP_MIN_LENGTH:
			nexus_particle_system_set_min_length(self, g_value_get_double(value));
			break;

		case PROP_MAX_DELAY:
			nexus_particle_system_set_max_delay(self, g_value_get_uint(value));
			break;

		case PROP_GLOW_TYPE:
			/* nexus_particle_system_set_glow_type(self, g_value_get_enum(value)); */
			break;

		case PROP_COLOR_1:
		case PROP_COLOR_2:
		case PROP_COLOR_3:
		case PROP_COLOR_4:
			nexus_particle_system_set_color(self,
			                                property_id - PROP_COLOR_1,
			                                g_value_get_boxed(value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
nexus_particle_system_get_property(GObject *object,
                                   guint property_id,
                                   GValue *value,
                                   GParamSpec *pspec)
{
	NexusParticleSystem *self = NEXUS_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_PULSE_COUNT:
			g_value_set_uint(value, nexus_particle_system_get_pulse_count(self));
			break;

		case PROP_PULSE_SIZE:
			g_value_set_uint(value, nexus_particle_system_get_pulse_size(self));
			break;

		case PROP_MAX_LENGTH:
			g_value_set_double(value, nexus_particle_system_get_max_length(self));
			break;

		case PROP_MIN_LENGTH:
			g_value_set_double(value, nexus_particle_system_get_min_length(self));
			break;

		case PROP_MAX_DELAY:
			g_value_set_uint(value, nexus_particle_system_get_max_delay(self));
			break;

		case PROP_GLOW_TYPE:
			/* g_value_set_enum(value, nexus_particle_system_get_glow_type(self)); */
			break;

		case PROP_COLOR_1:
		case PROP_COLOR_2:
		case PROP_COLOR_3:
		case PROP_COLOR_4:
		{
			GdkRGBA rgba;
			nexus_particle_system_get_color(self,
			                                property_id - PROP_COLOR_1,
			                                &rgba);
			g_value_set_boxed(value, &rgba);
			break;
		}

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static inline void
nexus_particle_system_init_pulse(NexusParticleSystem *self, Pulse *pulse)
{
	pulse->length = rand2f(self->priv->min_length, self->priv->max_length);
	pulse->color = rand() % 4;
	pulse->delay = rand() % self->priv->max_delay;

	if(randf() > 0.5f)
	{
		pulse->vy = 0.0f;
		pulse->y = randf();

		if(randf() > 0.5f)
		{
			/* Left to right */
			pulse->x = 0.0f;
			pulse->vx = pulse->length;
		}
		else
		{
			/* Right to left */
			pulse->x = 1.0f;
			pulse->vx = -pulse->length;
		}
	}
	else
	{
		pulse->vx = 0.0f;
		pulse->x = randf();

		if(randf() > 0.5f)
		{
			/* Top to bottom */
			pulse->y = 1.0f;
			pulse->vy = -pulse->length;
		}
		else
		{
			/* Bottom to top */
			pulse->y = 0.0f;
			pulse->vy = pulse->length;
		}
	}
}

void
nexus_particle_system_set_pulse_count(NexusParticleSystem *self, guint count)
{
	g_array_set_size(self->priv->pulses, count);

	/* Append pulses */
	if(self->priv->pulse_count < count)
	{
		guint i;

		for(i = self->priv->pulse_count; i < count; i++)
		{
			Pulse *pulse = &g_array_index(self->priv->pulses, Pulse, i);
			nexus_particle_system_init_pulse(self, pulse);
		}
	}

	self->priv->pulse_count = count;
}

guint
nexus_particle_system_get_pulse_count(NexusParticleSystem *self)
{
	return self->priv->pulse_count;
}

void
nexus_particle_system_set_pulse_size(NexusParticleSystem *self, guint size)
{
	self->priv->pulse_size = size;
}

guint
nexus_particle_system_get_pulse_size(NexusParticleSystem *self)
{
	return self->priv->pulse_size;
}

static inline void
nexus_particle_system_reinit_pulses(NexusParticleSystem *self)
{
	guint i;

	for(i = 0; i < self->priv->pulse_count; i++)
	{
		Pulse *pulse = &g_array_index(self->priv->pulses, Pulse, i);
		nexus_particle_system_init_pulse(self, pulse);
	}
}

void
nexus_particle_system_set_max_length(NexusParticleSystem *self, double length)
{
	self->priv->max_length = length;
	nexus_particle_system_reinit_pulses(self);
}

double
nexus_particle_system_get_max_length(NexusParticleSystem *self)
{
	return self->priv->max_length;
}

void
nexus_particle_system_set_min_length(NexusParticleSystem *self, double length)
{
	self->priv->min_length = length;
	nexus_particle_system_reinit_pulses(self);
}

double
nexus_particle_system_get_min_length(NexusParticleSystem *self)
{
	return self->priv->min_length;
}

void
nexus_particle_system_set_max_delay(NexusParticleSystem *self, guint delay)
{
	self->priv->max_delay = delay;
}

guint
nexus_particle_system_get_max_delay(NexusParticleSystem *self)
{
	return self->priv->max_delay;
}

void
nexus_particle_system_set_texture_dir(NexusParticleSystem *self, gchar *texturedir)
{
	self->priv->texturedir = g_strdup(texturedir);

	/* Load trail texture? */
	if(self->priv->texturedir && !self->priv->trailTexture)
	{
		gchar *filename;
		GError *error = NULL;

		filename = g_build_filename(self->priv->texturedir, "trail.png", NULL);
		self->priv->trailTexture = lw_texture_new_from_file(filename, &error);
		if(error != NULL)
		{
			g_critical("Could not load trail texture: %s", error->message);
			g_error_free(error);
		}
		g_free(filename);
	}

	/* Load glow texture? */
	if(self->priv->texturedir && !self->priv->glowTexture)
	{
		nexus_particle_system_set_glow_type(self, self->priv->glow_type);
	}
}

gchar*
nexus_particle_system_get_texture_dir(NexusParticleSystem *self)
{
	return self->priv->texturedir;
}

void
nexus_particle_system_set_glow_type(NexusParticleSystem *self, guint type)
{
	if(self->priv->texturedir)
	{
		gchar *filename;
		GError *error = NULL;

		if(self->priv->glowTexture)
		{
			g_object_unref(self->priv->glowTexture);
			self->priv->glowTexture = NULL;
		}

		switch(type)
		{
			case NexusGlowTypeSquare:
				filename = g_build_filename(self->priv->texturedir, "glow-square.png", NULL);
				break;

			case NexusGlowTypeSpiral:
				filename = g_build_filename(self->priv->texturedir, "glow-spiral.png", NULL);
				break;

			case NexusGlowTypeConcentricCircles:
				filename = g_build_filename(self->priv->texturedir, "glow-concentric-cirles.png", NULL);
				break;

			case NexusGlowTypeRadial:
			default:
				filename = g_build_filename(self->priv->texturedir, "glow-radial.png", NULL);
				break;
		}

		self->priv->glowTexture = lw_texture_new_from_file(filename, &error);
		if(error != NULL)
		{
			g_critical("Could not load glow texture: %s", error->message);
			g_error_free(error);
		}
		g_free(filename);
	}

	self->priv->glow_type = type;
}

guint
nexus_particle_system_get_glow_type(NexusParticleSystem *self)
{
	return self->priv->glow_type;
}

void
nexus_particle_system_set_color(NexusParticleSystem *self, int index, GdkRGBA *color)
{
	g_return_if_fail(index > 0 || index < 3);

	self->priv->colors[index] = *color;
}

void
nexus_particle_system_get_color(NexusParticleSystem *self, int index, GdkRGBA *color)
{
	g_return_if_fail(index > 0 || index < 3);

	*color = self->priv->colors[index];
}

void
nexus_particle_system_update(NexusParticleSystem *self, int ms_since_last_paint)
{
	guint i;
	Pulse *pulse;

	for(i = 0; i < self->priv->pulse_count; i++)
	{
		pulse = &g_array_index(self->priv->pulses, Pulse, i);

		if(pulse->delay > 0)
		{
			pulse->delay -= ms_since_last_paint;
			continue;
		}

		if(pulse->vy == 0.0f)
		{
			pulse->x += pulse->vx * ms_since_last_paint / 1000;

			if(pulse->vx < 0 && pulse->x < pulse->vx)
				nexus_particle_system_init_pulse(self, pulse);
			if(pulse->vx > 0 && pulse->x > 1.0f + pulse->vx)
				nexus_particle_system_init_pulse(self, pulse);
		}
		else
		{
			pulse->y += pulse->vy * ms_since_last_paint / 1000;

			if(pulse->vy < 0 && pulse->y < pulse->vy)
				nexus_particle_system_init_pulse(self, pulse);
			if(pulse->vy > 0 && pulse->y > 1.0f + pulse->vy)
				nexus_particle_system_init_pulse(self, pulse);
		}
	}
}

/* Draws quad for horizontal trail. */
#define DRAW_QUAD_H(x1, y1, x2, y2)	\
	glBegin(GL_QUADS);				\
		glTexCoord2d(0.0f, 0.0f);	\
		glVertex2f(x1, y1);			\
		glTexCoord2d(0.0f, 1.0f);	\
		glVertex2f(x1, y2);			\
		glTexCoord2d(1.0f, 1.0f);	\
		glVertex2f(x2, y2);			\
		glTexCoord2d(1.0f, 0.0f);	\
		glVertex2f(x2, y1);			\
	glEnd()

/* Draws quad for vertical trail (Different texture coordinates than DRAW_QUAD_H). */
#define DRAW_QUAD_V(x1, y1, x2, y2)	\
	glBegin(GL_QUADS);				\
		glTexCoord2d(0.0f, 1.0f);	\
		glVertex2f(x1, y1);			\
		glTexCoord2d(1.0f, 1.0f);	\
		glVertex2f(x1, y2);			\
		glTexCoord2d(1.0f, 0.0f);	\
		glVertex2f(x2, y2);			\
		glTexCoord2d(0.0f, 0.0f);	\
		glVertex2f(x2, y1);			\
	glEnd()

void
nexus_particle_system_draw(NexusParticleSystem *self, int size)
{
	float half_glow_size = (4.0f * self->priv->pulse_size) / (2.0f * size);
	float half_pulse_size = self->priv->pulse_size / (2.0f * size);
	guint i;
	Pulse *pulse;

	for(i = 0; i < self->priv->pulse_count; i++)
	{
		pulse = &g_array_index(self->priv->pulses, Pulse, i);

		if(pulse->delay > 0) continue;

		/* Set color */
		glColor4f(self->priv->colors[pulse->color].red,
		          self->priv->colors[pulse->color].green,
		          self->priv->colors[pulse->color].blue,
		          self->priv->colors[pulse->color].alpha);

		/* Draw glow */
		lw_texture_enable(self->priv->glowTexture);
		DRAW_QUAD_H(pulse->x - half_glow_size, pulse->y - half_glow_size,
		            pulse->x + half_glow_size, pulse->y + half_glow_size);
		lw_texture_disable(self->priv->glowTexture);

		/* Draw trail */
		#define x1 (pulse->x - half_pulse_size)
		#define y1 (pulse->y - half_pulse_size)
		#define x2 (pulse->x + half_pulse_size)
		#define y2 (pulse->y + half_pulse_size)

		lw_texture_enable(self->priv->trailTexture);
		if(pulse->vx > 0)       /* Left to right */
		{
			DRAW_QUAD_H(x2, y2, x1 - pulse->length, y1);
		}
		else if(pulse->vx < 0)  /* Right to left */
		{
			DRAW_QUAD_H(x1, y1, x2 + pulse->length, y2);
		}
		else if(pulse->vy > 0)  /* Bottom to top */
		{
			DRAW_QUAD_V(x2, y2, x1, y1 - pulse->length);
		}
		else                    /* Top to bottom */
		{
			DRAW_QUAD_V(x1, y1, x2, y2 + pulse->length);
		}
		lw_texture_disable(self->priv->trailTexture);

		#undef x1
		#undef y1
		#undef x2
		#undef y2
	}
}

static void
nexus_particle_system_init(NexusParticleSystem *self)
{
    self->priv = NEXUS_PARTICLE_SYSTEM_GET_PRIVATE(self);

    self->priv->pulse_count = 0;
    self->priv->pulse_size = 8;

    self->priv->max_length = 0.4;
    self->priv->min_length = 0.1;

    self->priv->max_delay = 5000;

    self->priv->pulses = g_array_new(FALSE, TRUE, sizeof(Pulse));

	self->priv->texturedir = NULL;
	self->priv->glow_type = NexusGlowTypeRadial;
    self->priv->glowTexture = NULL;
    self->priv->trailTexture = NULL;
}

static void
nexus_particle_system_dispose(GObject *object)
{
	NexusParticleSystem *self = NEXUS_PARTICLE_SYSTEM(object);

	if(self->priv->glowTexture)
	{
		g_object_unref(self->priv->glowTexture);
		self->priv->glowTexture = NULL;
	}

	if(self->priv->trailTexture)
	{
		g_object_unref(self->priv->trailTexture);
		self->priv->trailTexture = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(nexus_particle_system_parent_class)->dispose(object);
}

static void
nexus_particle_system_finalize(GObject *object)
{
	NexusParticleSystem *self = NEXUS_PARTICLE_SYSTEM(object);

	g_array_free(self->priv->pulses, TRUE);
	g_free(self->priv->texturedir);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(nexus_particle_system_parent_class)->finalize(object);
}

static void
nexus_particle_system_class_init(NexusParticleSystemClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = nexus_particle_system_set_property;
	gobject_class->get_property = nexus_particle_system_get_property;
	gobject_class->dispose = nexus_particle_system_dispose;
	gobject_class->finalize = nexus_particle_system_finalize;

	g_type_class_add_private(klass, sizeof(NexusParticleSystemPrivate));

	obj_properties[PROP_PULSE_COUNT] =
		g_param_spec_uint("pulse-count",
		                  "Number of pulses",
		                  "Number of pulses",
		                  0,
		                  300,
		                  0,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_PULSE_SIZE] =
		g_param_spec_uint("pulse-size",
		                  "Pulse size",
		                  "Size of a pulse",
		                  8,
		                  64,
		                  16,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_MAX_LENGTH] =
		g_param_spec_double("max-length",
		                    "Max length",
		                    "Max length of a pulse",
		                    0.0,
		                    1.0,
		                    0.4,
		                    G_PARAM_READWRITE);

	obj_properties[PROP_MIN_LENGTH] =
		g_param_spec_double("min-length",
		                    "Min length",
		                    "Min length of a pulse",
		                    0.0,
		                    1.0,
		                    0.1,
		                    G_PARAM_READWRITE);

	obj_properties[PROP_MAX_DELAY] =
		g_param_spec_uint("max-delay",
		                  "Max delay",
		                  "Max delay of a pulse in milliseconds",
		                  1,
		                  10000,
		                  5000,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_GLOW_TYPE] =
		g_param_spec_uint("glow-type",
		                  "Glow type",
		                  "The shape of the glow",
		                  0,
		                  3,
		                  NexusGlowTypeRadial,
		                  G_PARAM_WRITABLE);

	obj_properties[PROP_COLOR_1] =
		g_param_spec_boxed("color1",
		                   "Color 1",
		                   "Pulse color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	obj_properties[PROP_COLOR_2] =
		g_param_spec_boxed("color2",
		                   "Color 2",
		                   "Pulse color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	obj_properties[PROP_COLOR_3] =
		g_param_spec_boxed("color3",
		                   "Color 3",
		                   "Pulse color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	obj_properties[PROP_COLOR_4] =
		g_param_spec_boxed("color4",
		                   "Color 4",
		                   "Pulse color",
		                   GDK_TYPE_RGBA,
		                   G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

