/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Fyrmir@gmx.de>
 *
 */

#include "config.h"

#include <string.h>
#include <sys/time.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <gio/gio.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <libpeas/peas.h>

#include <GL/glew.h>
#include <GL/glxew.h>

#include <livewallpaper/core.h>

#include "application.h"


#define LW_APPLICATION_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), LW_TYPE_APPLICATION, LwApplicationPrivate))

struct _LwApplicationPrivate
{
	Display *dpy;
	GdkWindow *win;
	GLXContext glc;

	GSettings *settings;

	gboolean active;

	PeasEngine *peas_engine;
	PeasExtension *active_plugin;
	PeasPluginInfo *active_plugin_info;

	gint n_outputs;
	GList *outputs;

	gboolean done;
	guint ms_per_iteration;

	LwCairoTexture *fps_texture;

	/* Command line options */
	gboolean show_fps;
	gboolean lower;
};

enum
{
	PROP_0,

	PROP_ACTIVE,
	PROP_ACTIVE_PLUGIN,
	PROP_FPS_LIMIT,

	N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(LwApplication, lw_application, G_TYPE_OBJECT)


static void
lw_application_load_active_plugin(LwApplication *self, const gchar *name)
{
	PeasEngine *engine = self->priv->peas_engine;
	PeasExtension *plugin;
	PeasPluginInfo *info = peas_engine_get_plugin_info(engine, name);

	if(info == NULL)
	{
		g_critical("Could not find plugin %s", name);
		return;
	}

	peas_engine_load_plugin(engine, info);
	if(!peas_engine_provides_extension(engine, info, LW_TYPE_LIVEWALLPAPER_PLUGIN))
	{
		g_critical("Could not load plugin: %s is not a livewallpaper plugin", name);
		return;
	}
	plugin = peas_engine_create_extension(engine, info, LW_TYPE_LIVEWALLPAPER_PLUGIN, NULL);

	lw_livewallpaper_plugin_init_plugin(LW_LIVEWALLPAPER_PLUGIN(plugin));

	self->priv->active_plugin = plugin;
	self->priv->active_plugin_info = info;
}

static void
lw_application_unload_active_plugin(LwApplication *self)
{
	PeasEngine *engine = self->priv->peas_engine;

	if(!self->priv->active_plugin_info) return;

	g_object_unref(self->priv->active_plugin);
	peas_engine_unload_plugin(engine, self->priv->active_plugin_info);

	self->priv->active_plugin = NULL;
	self->priv->active_plugin_info = NULL;
}

static void
lw_application_set_property(GObject *object,
                       guint property_id,
                       const GValue *value,
                       GParamSpec *pspec)
{
	LwApplication *self = LW_APPLICATION(object);

	switch(property_id)
	{
		case PROP_ACTIVE:
			self->priv->active = g_value_get_boolean(value);
			break;

		case PROP_ACTIVE_PLUGIN:
		{
			const gchar *name = g_value_get_string(value);

			/* Restore viewport for the old plugin? */
			if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
			{
				lw_livewallpaper_plugin_restore_viewport(
					LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin));
			}

			/* Unload old and load new plugin */
			lw_application_unload_active_plugin(self);
			lw_application_load_active_plugin(self, name);

			/* Adjust viewport for the new plugin? */
			if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
			{
				LwOutput *o = self->priv->outputs->data;
				lw_livewallpaper_plugin_adjust_viewport(
					LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), o);
			}

			break;
		}

		case PROP_FPS_LIMIT:
		{
			guint fps = g_value_get_uint(value);
			self->priv->ms_per_iteration = (fps == 0) ? 0 : 1000 / fps;
			break;
		}

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
lw_application_get_property(GObject *object,
                       guint property_id,
                       GValue *value,
                       GParamSpec *pspec)
{
	LwApplication *self = LW_APPLICATION(object);

	switch(property_id)
	{
		case PROP_ACTIVE:
			g_value_set_boolean(value, self->priv->active);
			break;

		case PROP_ACTIVE_PLUGIN:
			if(self->priv->active_plugin_info)
				g_value_set_string(value,
					peas_plugin_info_get_name(self->priv->active_plugin_info));
			else
				g_value_set_string(value, "");
			break;

		case PROP_FPS_LIMIT:
			g_value_set_uint(value, 1000 / self->priv->ms_per_iteration);
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static GOptionEntry options[] = {
	{ "show-fps", 'f', 0, G_OPTION_ARG_NONE, NULL, N_("Show FPS in the lower right corner"),               NULL },
	{ "lower"   ,  0 , 0, G_OPTION_ARG_NONE, NULL, N_("Move the window behind all other desktop windows"), NULL},
	{ NULL }
};

static gboolean
lw_application_parse_command_line(LwApplication *self, gint argc, gchar **argv)
{
	GError *error = NULL;
	GOptionContext *context;

	/* Set arg_data of all options */
	options[0].arg_data = &self->priv->show_fps;
	options[1].arg_data = &self->priv->lower;

	context = g_option_context_new(NULL);
	g_option_context_add_main_entries(context, options, "livewallpaper");

	if(!g_option_context_parse(context, &argc, &argv, &error))
	{
		g_critical("Could not parse command line: %s\n", error->message);
		g_option_context_free(context);
		return FALSE;
	}

	g_option_context_free(context);
	return TRUE;
}


/************* OpenGL Window *************/
static gboolean
lw_application_create_opengl_desktop_window(LwApplication *self)
{
	GLint gl_attr[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
	XVisualInfo *vi;
	GdkWindowAttr attr;
	GdkWindow *root = gdk_get_default_root_window();

	attr.event_mask = 0;
	attr.width = gdk_screen_width() + 1; /* WORKAROUND */
	attr.height = gdk_screen_height();
	attr.wclass = GDK_INPUT_OUTPUT;

	attr.visual = NULL;
	self->priv->dpy = GDK_DISPLAY_XDISPLAY( gdk_display_get_default() );
	vi = glXChooseVisual(self->priv->dpy, 0, gl_attr);
	if(vi != NULL)
		attr.visual = gdk_x11_screen_lookup_visual(gdk_screen_get_default(), vi->visualid);

	if(attr.visual == NULL)
	{
		g_critical("Could not find matching visual");
		return FALSE;
	}

	attr.window_type = GDK_WINDOW_TOPLEVEL;
	attr.type_hint = GDK_WINDOW_TYPE_HINT_DESKTOP;

	self->priv->win = gdk_window_new(root, &attr, GDK_WA_VISUAL | GDK_WA_TYPE_HINT);

	gdk_window_show(self->priv->win);
	
	if(self->priv->lower)
		gdk_window_lower(self->priv->win);
	else
		gdk_window_raise(self->priv->win);

	/* Create OpenGL context */
	self->priv->glc = glXCreateContext(self->priv->dpy, vi, NULL, GL_TRUE);
	if(self->priv->glc == NULL)
	{
		gdk_window_destroy(self->priv->win);
		XFree(vi);

		g_critical("Could not create gl context");
		return FALSE;
	}

	glXMakeCurrent(self->priv->dpy, GDK_WINDOW_XID(self->priv->win), self->priv->glc);

	XFree(vi);
	return TRUE;
}

static void
lw_application_destroy_opengl_desktop_window(LwApplication *self)
{
	glXMakeCurrent(self->priv->dpy, None, NULL);
	glXDestroyContext(self->priv->dpy, self->priv->glc);
	gdk_window_destroy(self->priv->win);
}


/************* Output Handling *************/
static void
lw_application_init_outputs(LwApplication *self)
{
	GdkScreen *screen = gdk_screen_get_default();
	gint i, n_monitors = gdk_screen_get_n_monitors(screen);
	GdkRectangle g1, g2;

	if(g_settings_get_enum(self->priv->settings, "multioutput-mode") == 1)
	{
		/* Multiple wallpapers */
		for(i = 0; i < n_monitors; i++)
		{
			gint j;
			gboolean found = FALSE;
			
			gdk_screen_get_monitor_geometry(screen, i, &g1);
			
			for(j = 0; j < i; j++)
			{
				gdk_screen_get_monitor_geometry(screen, j, &g2);

				if(gdk_rectangle_intersect(&g1, &g2, NULL))
				{
					found = TRUE;
					break;
				}
			}

			if(!found)
			{
				/* Append new output */
				LwOutput *o = g_object_new(LW_TYPE_OUTPUT,
					                       "id", i,
					                       "x", g1.x,
					                       "y", g1.y,
					                       "width", g1.width,
					                       "height", g1.height,
					                       NULL);

				self->priv->outputs = g_list_prepend(self->priv->outputs, o);
			}
		}
	}
	else
	{
		LwOutput *o;

		/* One big wallpaper */
		g1.x = g1.y = g1.width = g1.height = 0;
		for(i = 0; i < n_monitors; i++)
		{
			gdk_screen_get_monitor_geometry(screen, i, &g2);
			gdk_rectangle_union(&g1, &g2, &g1);
		}

		/* Append new output */
		o = g_object_new(LW_TYPE_OUTPUT,
		                 "id", i,
		                 "x", g1.x,
		                 "y", g1.y,
		                 "width", g1.width,
		                 "height", g1.height,
		                 NULL);

		self->priv->outputs = g_list_prepend(self->priv->outputs, o);
	}

	self->priv->n_outputs = g_list_length(self->priv->outputs);
}

static void
lw_application_finalize_outputs(LwApplication *self)
{
	g_list_free_full(self->priv->outputs, g_object_unref);
	self->priv->n_outputs = 0;
	self->priv->outputs = NULL;
}

static void
lw_application_update_outputs(LwApplication *self)
{
	/* Restore viewport? */
	if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
	{
		lw_livewallpaper_plugin_restore_viewport(
			LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin));
	}

	/* Update outputs */
	lw_application_finalize_outputs(self);
	lw_application_init_outputs(self);

	/* Adjust viewport? */
	if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
	{
		LwOutput *o = self->priv->outputs->data;

		lw_output_set_viewport(o);

		lw_livewallpaper_plugin_adjust_viewport(
			LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), o);
	}
}

static void
lw_application_update_outputs_from_screen(GdkScreen *screen, LwApplication *self)
{
	(void) screen;

	lw_application_update_outputs(self);
}

static void
lw_application_update_outputs_from_settings(GSettings *settings, gchar *key, LwApplication *self)
{
	(void) settings;
	(void) key;

	lw_application_update_outputs(self);
}

static void
lw_application_resize_window(GdkScreen *screen, LwApplication *self)
{
	gdk_window_resize(self->priv->win,
		gdk_screen_get_width(screen) + 1, /* WORKAROUND */
		gdk_screen_get_height(screen));
}


/************* FPS Helpers *************/
static void
lw_application_finalize_fps(LwApplication *self)
{
	if(self->priv->fps_texture == NULL) return;

	g_object_unref(self->priv->fps_texture);
	self->priv->fps_texture = NULL;
}

static void
lw_application_init_fps(LwApplication *self)
{
	GError *error = NULL;

	self->priv->fps_texture = lw_cairo_texture_new(256, 256, &error);
	if(error != NULL)
	{
		g_warning("Could not create FPS texture: %s", error->message);
		g_error_free(error);
		error = NULL;
	}
}

static int
lw_application_get_fps(int ms_since_last_paint, gboolean get_only)
{
	#define DATA_COUNT 100
	static int index = 0;
	static int sum = 0;
	static int data[DATA_COUNT] = {0};

	if(!get_only)
	{
		sum -= data[index];
		sum += ms_since_last_paint;
		data[index] = ms_since_last_paint;

		index++;
		if(index == DATA_COUNT) index = 0;
	}

	if(sum == 0) return 0;

	return (DATA_COUNT * 1000) / sum;
}

static void
lw_application_show_fps(LwApplication *self, LwOutput *output)
{
	static gint last_fps = -1;
	static guint text_width, text_height;
	guint width = lw_output_get_width(output),
	      height = lw_output_get_height(output);
	gint fps = lw_application_get_fps(0, TRUE);
	gboolean was_blending_enabled = FALSE;
	LwTextureMatrix m;

	if(self->priv->fps_texture == NULL) return;

	/* Update FPS texture? */
	if(fps != last_fps)
	{
		cairo_t *cr = lw_cairo_texture_cairo_create(self->priv->fps_texture);
		cairo_text_extents_t extents;
		gchar s[16];

		sprintf(s, "%d FPS", fps);

		/* Select font */
		cairo_select_font_face(cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_BOLD);
		cairo_set_font_size(cr, 20);

		/* Get size of text */
		cairo_text_extents(cr, s, &extents);
		text_width = 20 + extents.width;
		text_height = 20 + extents.height;

		if(text_width > 256) text_width = 256;
		if(text_height > 256) text_height = 256;

		/* Fill background */
		cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
		cairo_rectangle(cr, 0, 0, text_width, text_height);
		cairo_fill(cr);

		/* Show text */
		cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
		cairo_move_to(cr, 10 - extents.x_bearing, 10 + extents.height);
		cairo_show_text(cr, s);

		cairo_destroy(cr);
		lw_cairo_texture_update(self->priv->fps_texture);
		last_fps = fps;
	}

	/* Save and reset the coordinate system */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glOrtho(0, width, height, 0, -1.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	/* Show FPS */
	glColor3f(1.0f, 1.0f, 1.0f);

	if(glIsEnabled(GL_BLEND)) was_blending_enabled = TRUE;
	glDisable(GL_BLEND);
	lw_texture_enable(LW_TEXTURE(self->priv->fps_texture));

	m = LW_TEXTURE(self->priv->fps_texture)->matrix;

	glBegin(GL_QUADS);
		glTexCoord2f(
			LW_TEX_COORD_X(m, 0.0f),
			LW_TEX_COORD_Y(m, 0.0f)
		);
		glVertex2f(width - text_width, height - text_height);

		glTexCoord2f(
			LW_TEX_COORD_X(m, 0.0f),
			LW_TEX_COORD_Y(m, text_height)
		);
		glVertex2f(width - text_width, height);

		glTexCoord2f(
			LW_TEX_COORD_X(m, text_width),
			LW_TEX_COORD_Y(m, text_height)
		);
		glVertex2f(width, height);

		glTexCoord2f(
			LW_TEX_COORD_X(m, text_width),
			LW_TEX_COORD_Y(m, 0.0f)
		);
		glVertex2f(width, height - text_height);
	glEnd();

	lw_texture_disable(LW_TEXTURE(self->priv->fps_texture));
	if(was_blending_enabled) glEnable(GL_BLEND);

	/* Restore the coordinate system */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}


static inline int
timeval_diff(struct timeval *tv1, struct timeval *tv2)
{
	/* This function is completely taken from compiz. */

	if(tv1->tv_sec == tv2->tv_sec || tv1->tv_usec >= tv2->tv_usec)
		return (((tv1->tv_sec - tv2->tv_sec) * 1000000) + (tv1->tv_usec - tv2->tv_usec)) / 1000;
	else
		return (((tv1->tv_sec - 1 - tv2->tv_sec) * 1000000) + (1000000 + tv1->tv_usec - tv2->tv_usec)) / 1000;
}

int
lw_application_run(LwApplication *self, gint argc, gchar **argv)
{
	GLenum error;
	LwOutput *o;
	struct timeval last_paint;

	/* Parse command line */
	if(!lw_application_parse_command_line(self, argc, argv)) return 1;

	/* Create desktop window */
	if(!lw_application_create_opengl_desktop_window(self))	return 1;

	/* Initialize GLEW */
	error = glewInit();
	if(error != GLEW_OK)
	{
		g_critical("Could not initialize glew: %s", glewGetErrorString(error));
		return 1;
	}

	/* Bind settings to properties */
	g_settings_bind(self->priv->settings, "active", self, "active", G_SETTINGS_BIND_GET);
	g_settings_bind(self->priv->settings, "active-plugin", self, "active-plugin", G_SETTINGS_BIND_GET);
	g_settings_bind(self->priv->settings, "fps-limit", self, "fps-limit", G_SETTINGS_BIND_GET);

	/* Connect signals */
	g_signal_connect(gdk_screen_get_default(), "monitors-changed",
		G_CALLBACK(lw_application_update_outputs_from_screen), self);
	g_signal_connect(gdk_screen_get_default(), "size-changed",
		G_CALLBACK(lw_application_resize_window), self);
	g_signal_connect(self->priv->settings, "changed::multioutput-mode",
		G_CALLBACK(lw_application_update_outputs_from_settings), self);

	/* Initialize OpenGL */
	glEnable(GL_SCISSOR_TEST);
	glClearColor(0.0, 0.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* Initialize FPS font */
	lw_application_init_fps(self);

	/* Initialize outputs */
	lw_application_init_outputs(self);

	/* Main loop */
	if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
	{
		/* Get current output */
		o = self->priv->outputs->data;

		/* Adjust viewport */
		lw_output_set_viewport(o);

		lw_livewallpaper_plugin_adjust_viewport(
			LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), o);
	}

	gettimeofday(&last_paint, 0);

	while(!self->priv->done)
	{
		struct timeval start;
		gint ms_since_last_paint;
		
		gettimeofday(&start, 0);
		ms_since_last_paint = timeval_diff(&start, &last_paint);

		if(self->priv->active && self->priv->active_plugin_info)
		{
			/* Paint all outputs */
			GList *outputs = self->priv->outputs;
			
			/* Prepare paint */
			lw_livewallpaper_plugin_prepare_paint(
				LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), ms_since_last_paint);

			while(outputs)
			{
				if(self->priv->n_outputs > 1)
				{
					/* Get current output */
					o = outputs->data;

					/* Adjust viewport */
					lw_output_set_viewport(o);

					lw_livewallpaper_plugin_adjust_viewport(
						LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), o);
				}
				else
				{
					/* Get current output */
					o = self->priv->outputs->data;
				}

				/* Paint */
				lw_livewallpaper_plugin_paint(
					LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin), o);

				/* Show FPS? */
				if(self->priv->show_fps)
					lw_application_show_fps(self, o);

				if(self->priv->n_outputs > 1)
				{
					/* Restore viewport */
					lw_livewallpaper_plugin_restore_viewport(
						LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin));

					o = NULL;
				}

				outputs = outputs->next;
			}

			/* Done paint */
			lw_livewallpaper_plugin_done_paint(
				LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin));
		}
		else
		{
			/* Clear window because there is no active livewallpaper plugin */
			glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
			glClear(GL_COLOR_BUFFER_BIT);
		}

		glXSwapBuffers(self->priv->dpy, GDK_WINDOW_XID(self->priv->win));
		last_paint = start;

		/* Handle events */
		g_main_context_iteration(NULL, FALSE);

		/* Measure FPS */
		if(self->priv->show_fps)
		{
			lw_application_get_fps(ms_since_last_paint, FALSE);
		}

		/* Limit FPS */
		if(self->priv->ms_per_iteration > 0)
		{
			struct timeval end;
			gint ms_to_sleep;
			
			gettimeofday(&end, 0);
			ms_to_sleep = self->priv->ms_per_iteration - timeval_diff(&end, &start);

			if(ms_to_sleep > 0) g_usleep(ms_to_sleep * 1000);
		}
	}

	if(self->priv->active_plugin_info && self->priv->n_outputs == 1)
	{
		/* Restore viewport */
		lw_livewallpaper_plugin_restore_viewport(
			LW_LIVEWALLPAPER_PLUGIN(self->priv->active_plugin));
	}

	/* Exit application */
	lw_application_finalize_fps(self);
	lw_application_finalize_outputs(self);
	lw_application_destroy_opengl_desktop_window(self);

	return 0;
}

void
lw_application_quit(LwApplication *self)
{
	self->priv->done = TRUE;
}

static void
lw_application_init(LwApplication *self)
{
	PeasEngine *engine = peas_engine_get_default();
	gchar *plugindir, *plugindatadir;

	self->priv = LW_APPLICATION_GET_PRIVATE(self);

	/* Initialize settings */
	self->priv->settings = g_settings_new("net.launchpad.livewallpaper");

	/* Initialize plugin engine */
	peas_engine_enable_loader(engine, "vala");
	peas_engine_enable_loader(engine, "python");

	plugindir = g_build_filename(g_get_home_dir(), ".livewallpaper", "plugins", NULL);
	peas_engine_add_search_path(engine, plugindir, NULL);
	g_free(plugindir);

	plugindir = g_build_filename(LIBDIR, "livewallpaper", "plugins", NULL);
	plugindatadir = g_build_filename(DATADIR, "livewallpaper", "plugins", NULL);
	peas_engine_add_search_path(engine, plugindir, plugindatadir);
	g_free(plugindir);
	g_free(plugindatadir);

	self->priv->peas_engine = engine;

	/* Set other default values */
	self->priv->active = TRUE;
	self->priv->n_outputs = 0;
	self->priv->outputs = NULL;
	self->priv->done = FALSE;
	self->priv->ms_per_iteration = 20;
	self->priv->fps_texture = NULL;
	self->priv->show_fps = FALSE;
	self->priv->lower = FALSE;
}

static void
lw_application_class_init(LwApplicationClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = lw_application_set_property;
	gobject_class->get_property = lw_application_get_property;

	g_type_class_add_private(klass, sizeof(LwApplicationPrivate));

	obj_properties[PROP_ACTIVE] =
		g_param_spec_boolean("active",
		                     "Start or stop the livewallpaper",
		                     "Set whether the livewallpaper is running or not",
		                     TRUE,
		                     G_PARAM_READWRITE);

	obj_properties[PROP_ACTIVE_PLUGIN] =
		g_param_spec_string("active-plugin",
		                    "Active plugin",
		                    "The name of the active livewallpaper plugin",
		                    NULL,
		                    G_PARAM_READWRITE);

	obj_properties[PROP_FPS_LIMIT] =
		g_param_spec_uint("fps-limit",
		                  "FPS Limit",
		                  "Limit the FPS to this value",
		                  0,
		                  120,
		                  50,
		                  G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

